#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Fix unnamed constraints.

Revision ID: f8c799db4aa0
Revises: 10d2738b67df
Create Date: 2016-05-18 18:08:19.331412

"""

# revision identifiers, used by Alembic.
revision = 'f8c799db4aa0'
down_revision = '10d2738b67df'
import copy

from alembic import op
import six

from cloudkitty.rating.hash.db.sqlalchemy.alembic.models import (
    f8c799db4aa0_fix_unnamed_constraints as models)

OPS = {
    'foreignkey': {
        'hashmap_fields': [
            ('hashmap_fields_service_id_fkey',
             'fk_hashmap_fields_service_id_hashmap_services',
             {
                 'args': [
                     'hashmap_services',
                     ['service_id'],
                     ['id']],
                 'kwargs': {'ondelete': 'CASCADE'}})],
        'hashmap_thresholds': [
            ('hashmap_thresholds_field_id_fkey',
             'fk_hashmap_thresholds_field_id_hashmap_fields',
             {
                 'args': [
                     'hashmap_fields',
                     ['field_id'],
                     ['id']],
                 'kwargs': {'ondelete': 'CASCADE'}}),
            ('hashmap_thresholds_group_id_fkey',
             'fk_hashmap_thresholds_group_id_hashmap_groups',
             {
                 'args': [
                     'hashmap_groups',
                     ['group_id'],
                     ['id']],
                 'kwargs': {'ondelete': 'SET NULL'}}),
            ('hashmap_thresholds_service_id_fkey',
             'fk_hashmap_thresholds_service_id_hashmap_services',
             {
                 'args': [
                     'hashmap_services',
                     ['service_id'],
                     ['id']],
                 'kwargs': {'ondelete': 'CASCADE'}})],
        'hashmap_mappings': [
            ('hashmap_maps_field_id_fkey',
             'fk_hashmap_maps_field_id_hashmap_fields',
             {
                 'args': [
                     'hashmap_fields',
                     ['field_id'],
                     ['id']],
                 'kwargs': {'ondelete': 'CASCADE'}}),
            ('hashmap_maps_group_id_fkey',
             'fk_hashmap_maps_group_id_hashmap_groups',
             {
                 'args': [
                     'hashmap_groups',
                     ['group_id'],
                     ['id']],
                 'kwargs': {'ondelete': 'SET NULL'}}),
            ('hashmap_maps_service_id_fkey',
             'fk_hashmap_maps_service_id_hashmap_services',
             {
                 'args': [
                     'hashmap_fields',
                     ['field_id'],
                     ['id']],
                 'kwargs': {'ondelete': 'CASCADE'}})]
    },
    'primary': {
        'hashmap_services': [
            ('hashmap_services_pkey',
             'pk_hashmap_services',
             {'args': [['id']]})],
        'hashmap_fields': [
            ('hashmap_fields_pkey',
             'pk_hashmap_fields',
             {'args': [['id']]})],
        'hashmap_groups': [
            ('hashmap_groups_pkey',
             'pk_hashmap_groups',
             {'args': [['id']]})],
        'hashmap_mappings': [
            ('hashmap_maps_pkey',
             'pk_hashmap_maps',
             {'args': [['id']]})],
        'hashmap_thresholds': [
            ('hashmap_thresholds_pkey',
             'pk_hashmap_thresholds',
             {'args': [['id']]})]
    },
    'unique': {
        'hashmap_services': [
            ('hashmap_services_name_key',
             'uq_hashmap_services_name',
             {'args': [['name']]}),
            ('hashmap_services_service_id_key',
             'uq_hashmap_services_service_id',
             {'args': [['service_id']]})],
        'hashmap_fields': [
            ('hashmap_fields_field_id_key',
             'uq_hashmap_fields_field_id',
             {'args': [['field_id']]})],
        'hashmap_groups': [
            ('hashmap_groups_group_id_key',
             'uq_hashmap_groups_group_id',
             {'args': [['group_id']]}),
            ('hashmap_groups_name_key',
             'uq_hashmap_groups_name',
             {'args': [['name']]})],
        'hashmap_mappings': [
            ('hashmap_maps_mapping_id_key',
             'uq_hashmap_maps_mapping_id',
             {'args': [['mapping_id']]})],
        'hashmap_thresholds': [
            ('hashmap_thresholds_threshold_id_key',
             'uq_hashmap_thresholds_threshold_id',
             {'args': [['threshold_id']]})]}}

POST_OPS = {
    'primary': {
        'hashmap_mappings': [
            ('pk_hashmap_maps',
             'pk_hashmap_mappings',
             {'args': [['id']]})]
    }}


def upgrade_sqlite():
    # NOTE(sheeprine): Batch automatically recreates tables,
    # use it as a lazy way to recreate tables and transfer data automagically.
    for name, table in six.iteritems(models.Base.metadata.tables):
        with op.batch_alter_table(name, copy_from=table) as batch_op:
            # NOTE(sheeprine): Dummy operation to force recreate.
            # Easier than delete and create.
            batch_op.alter_column('id')


def upgrade_mysql():
    op.execute('SET FOREIGN_KEY_CHECKS=0;')
    tables = copy.deepcopy(models.Base.metadata.tables)
    # Copy first without constraints
    tables['hashmap_fields'].constraints = set()
    tables['hashmap_mappings'].constraints = set()
    tables['hashmap_thresholds'].constraints = set()
    for name, table in six.iteritems(tables):
        with op.batch_alter_table(name,
                                  copy_from=table,
                                  recreate='always') as batch_op:
            batch_op.alter_column('id')
    # Final copy with constraints
    for name, table in six.iteritems(models.Base.metadata.tables):
        with op.batch_alter_table(name,
                                  copy_from=table,
                                  recreate='always') as batch_op:
            batch_op.alter_column('id')
    op.execute('SET FOREIGN_KEY_CHECKS=1;')


def translate_op(op_, constraint_type, name, table, *args, **kwargs):
    if op_ == 'drop':
        op.drop_constraint(name, table, type_=constraint_type)
    else:
        if constraint_type == 'primary':
            func = op.create_primary_key
        elif constraint_type == 'unique':
            func = op.create_unique_constraint
        elif constraint_type == 'foreignkey':
            func = op.create_foreign_key
        func(name, table, *args, **kwargs)


def upgrade_postgresql():
    # NOTE(sheeprine): No automagic stuff here.
    # Check if tables need additional work
    conn = op.get_bind()
    res = conn.execute(
        "SELECT * FROM INFORMATION_SCHEMA.REFERENTIAL_CONSTRAINTS"
        " WHERE CONSTRAINT_NAME = 'hashmap_thresholds_field_id_fkey';")
    if res.rowcount:
        ops_list = [OPS, POST_OPS]
    else:
        ops_list = [POST_OPS]
    for cur_ops in ops_list:
        for constraint_type in ('foreignkey', 'unique', 'primary'):
            for table_name, constraints in six.iteritems(
                    cur_ops.get(constraint_type, dict())):
                for constraint in constraints:
                    old_name = constraint[0]
                    translate_op(
                        'drop',
                        constraint_type,
                        old_name,
                        table_name)
        for constraint_type in ('primary', 'unique', 'foreignkey'):
            for table_name, constraints in six.iteritems(
                    cur_ops.get(constraint_type, dict())):
                for constraint in constraints:
                    new_name = constraint[1]
                    params = constraint[2]
                    translate_op(
                        'create',
                        constraint_type,
                        new_name,
                        table_name,
                        *params.get('args', []),
                        **params.get('kwargs', {}))


def upgrade():
    dialect = op.get_context().dialect
    if dialect.name == 'sqlite':
        upgrade_sqlite()
    elif dialect.name == 'mysql':
        upgrade_mysql()
    elif dialect.name == 'postgresql':
        upgrade_postgresql()



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
from oslo_db.sqlalchemy import utils
import sqlalchemy

from cloudkitty import config  # NOQA
from cloudkitty import db
from cloudkitty.db import api
from cloudkitty.db.sqlalchemy import migration
from cloudkitty.db.sqlalchemy import models


def get_backend():
    return DBAPIManager


class State(api.State):

    def get_state(self, name):
        session = db.get_session()
        q = utils.model_query(
            models.StateInfo,
            session)
        q = q.filter(models.StateInfo.name == name)
        return q.value(models.StateInfo.state)

    def set_state(self, name, state):
        session = db.get_session()
        with session.begin():
            try:
                q = utils.model_query(
                    models.StateInfo,
                    session)
                q = q.filter(models.StateInfo.name == name)
                q = q.with_lockmode('update')
                db_state = q.one()
                db_state.state = state
            except sqlalchemy.orm.exc.NoResultFound:
                db_state = models.StateInfo(name=name,
                                            state=state)
                session.add(db_state)
        return db_state.state

    def get_metadata(self, name):
        session = db.get_session()
        q = utils.model_query(
            models.StateInfo,
            session)
        q.filter(models.StateInfo.name == name)
        return q.value(models.StateInfo.s_metadata)

    def set_metadata(self, name, metadata):
        session = db.get_session()
        with session.begin():
            try:
                q = utils.model_query(
                    models.StateInfo,
                    session)
                q = q.filter(models.StateInfo.name == name)
                q = q.with_lockmode('update')
                db_state = q.one()
                db_state.s_metadata = metadata
            except sqlalchemy.orm.exc.NoResultFound:
                db_state = models.StateInfo(name=name,
                                            s_metadata=metadata)
                session.add(db_state)


class ModuleInfo(api.ModuleInfo):
    """Base class for module info management."""

    def get_priority(self, name):
        session = db.get_session()
        q = utils.model_query(
            models.ModuleStateInfo,
            session)
        q = q.filter(models.ModuleStateInfo.name == name)
        res = q.value(models.ModuleStateInfo.priority)
        if res:
            return int(res)
        else:
            return 1

    def set_priority(self, name, priority):
        session = db.get_session()
        with session.begin():
            try:
                q = utils.model_query(
                    models.ModuleStateInfo,
                    session)
                q = q.filter(
                    models.ModuleStateInfo.name == name)
                q = q.with_lockmode('update')
                db_state = q.one()
                db_state.priority = priority
            except sqlalchemy.orm.exc.NoResultFound:
                db_state = models.ModuleStateInfo(name=name,
                                                  priority=priority)
                session.add(db_state)
        return int(db_state.priority)

    def get_state(self, name):
        session = db.get_session()
        try:
            q = utils.model_query(
                models.ModuleStateInfo,
                session)
            q = q.filter(models.ModuleStateInfo.name == name)
            res = q.value(models.ModuleStateInfo.state)
            return bool(res)
        except sqlalchemy.orm.exc.NoResultFound:
            return None

    def set_state(self, name, state):
        session = db.get_session()
        with session.begin():
            try:
                q = utils.model_query(
                    models.ModuleStateInfo,
                    session)
                q = q.filter(models.ModuleStateInfo.name == name)
                q = q.with_lockmode('update')
                db_state = q.one()
                db_state.state = state
            except sqlalchemy.orm.exc.NoResultFound:
                db_state = models.ModuleStateInfo(name=name, state=state)
                session.add(db_state)
        return bool(db_state.state)


class ServiceToCollectorMapping(object):
    """Base class for service to collector mapping."""

    def get_mapping(self, service):
        session = db.get_session()
        try:
            q = utils.model_query(
                models.ServiceToCollectorMapping,
                session)
            q = q.filter(
                models.ServiceToCollectorMapping.service == service)
            return q.one()
        except sqlalchemy.orm.exc.NoResultFound:
            raise api.NoSuchMapping(service)

    def set_mapping(self, service, collector):
        session = db.get_session()
        with session.begin():
            try:
                q = utils.model_query(
                    models.ServiceToCollectorMapping,
                    session)
                q = q.filter(
                    models.ServiceToCollectorMapping.service == service)
                q = q.with_lockmode('update')
                db_mapping = q.one()
                db_mapping.collector = collector
            except sqlalchemy.orm.exc.NoResultFound:
                model = models.ServiceToCollectorMapping
                db_mapping = model(service=service,
                                   collector=collector)
                session.add(db_mapping)
        return db_mapping

    def list_services(self, collector=None):
        session = db.get_session()
        q = utils.model_query(
            models.ServiceToCollectorMapping,
            session)
        if collector:
            q = q.filter(
                models.ServiceToCollectorMapping.collector == collector)
        res = q.distinct().values(
            models.ServiceToCollectorMapping.service)
        return res

    def list_mappings(self, collector=None):
        session = db.get_session()
        q = utils.model_query(
            models.ServiceToCollectorMapping,
            session)
        if collector:
            q = q.filter(
                models.ServiceToCollectorMapping.collector == collector)
        res = q.all()
        return res

    def delete_mapping(self, service):
        session = db.get_session()
        q = utils.model_query(
            models.ServiceToCollectorMapping,
            session)
        q = q.filter(models.ServiceToCollectorMapping.service == service)
        r = q.delete()
        if not r:
            raise api.NoSuchMapping(service)


class DBAPIManager(object):

    @staticmethod
    def get_state():
        return State()

    @staticmethod
    def get_module_info():
        return ModuleInfo()

    @staticmethod
    def get_service_to_collector_mapping():
        return ServiceToCollectorMapping()

    @staticmethod
    def get_migration():
        return migration



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
from oslo_config import cfg
from stevedore import extension

from cloudkitty import config  # noqa
from cloudkitty.db import api as db_api
from cloudkitty import service

CONF = cfg.CONF
PROCESSORS_NAMESPACE = 'cloudkitty.rating.processors'


class ModuleNotFound(Exception):
    def __init__(self, name):
        self.name = name
        super(ModuleNotFound, self).__init__(
            "Module %s not found" % name)


class MultipleModulesRevisions(Exception):
    def __init__(self, revision):
        self.revision = revision
        super(MultipleModulesRevisions, self).__init__(
            "Can't apply revision %s to multiple modules." % revision)


class DBCommand(object):

    def __init__(self):
        self.rating_models = {}
        self._load_rating_models()

    def _load_rating_models(self):
        extensions = extension.ExtensionManager(
            PROCESSORS_NAMESPACE)
        self.rating_models = {}
        for ext in extensions:
            if hasattr(ext.plugin, 'db_api'):
                self.rating_models[ext.name] = ext.plugin.db_api

    def get_module_migration(self, name):
        if name == 'cloudkitty':
            mod_migration = db_api.get_instance().get_migration()
        else:
            try:
                module = self.rating_models[name]
                mod_migration = module.get_migration()
            except KeyError:
                raise ModuleNotFound(name)
        return mod_migration

    def get_migrations(self, name=None):
        if not name:
            migrations = []
            migrations.append(self.get_module_migration('cloudkitty'))
            for model in self.rating_models.values():
                migrations.append(model.get_migration())
            return migrations
        else:
            return [self.get_module_migration(name)]

    def check_revsion(self, revision):
        revision = revision or 'head'
        if revision not in ('base', 'head'):
            raise MultipleModulesRevisions(revision)

    def _version_change(self, cmd):
        revision = CONF.command.revision
        module = CONF.command.module
        if not module:
            self.check_revsion(revision)
        migrations = self.get_migrations(module)
        for migration in migrations:
            func = getattr(migration, cmd)
            func(revision)

    def upgrade(self):
        self._version_change('upgrade')

    def revision(self):
        migration = self.get_module_migration(CONF.command.module)
        migration.revision(CONF.command.message, CONF.command.autogenerate)

    def stamp(self):
        migration = self.get_module_migration(CONF.command.module)
        migration.stamp(CONF.command.revision)

    def version(self):
        migration = self.get_module_migration(CONF.command.module)
        migration.version()


def add_command_parsers(subparsers):
    command_object = DBCommand()

    parser = subparsers.add_parser('upgrade')
    parser.set_defaults(func=command_object.upgrade)
    parser.add_argument('--revision', nargs='?')
    parser.add_argument('--module', nargs='?')

    parser = subparsers.add_parser('stamp')
    parser.set_defaults(func=command_object.stamp)
    parser.add_argument('--revision', nargs='?')
    parser.add_argument('--module', required=True)

    parser = subparsers.add_parser('revision')
    parser.set_defaults(func=command_object.revision)
    parser.add_argument('-m', '--message')
    parser.add_argument('--autogenerate', action='store_true')
    parser.add_argument('--module', required=True)

    parser = subparsers.add_parser('version')
    parser.set_defaults(func=command_object.version)
    parser.add_argument('--module', required=True)


command_opt = cfg.SubCommandOpt('command',
                                title='Command',
                                help='Available commands',
                                handler=add_command_parsers)

CONF.register_cli_opt(command_opt)


def main():
    service.prepare_service()
    CONF.command.func()



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
from stevedore import enabled


class EnabledExtensionManager(enabled.EnabledExtensionManager):
    """CloudKitty Rating processor manager

    Override default EnabledExtensionManager to check for an internal
    object property in the extension.
    """

    def __init__(self, namespace, invoke_args=(), invoke_kwds={}):

        def check_enabled(ext):
            """Check if extension is enabled.

            """
            return ext.obj.enabled

        super(EnabledExtensionManager, self).__init__(
            namespace=namespace,
            check_func=check_enabled,
            invoke_on_load=True,
            invoke_args=invoke_args,
            invoke_kwds=invoke_kwds,
        )



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Gauvain Pocentek
#
import datetime

from cloudkitty import state
from cloudkitty import tests


class DBStateManagerTest(tests.TestCase):
    def setUp(self):
        super(DBStateManagerTest, self).setUp()
        self.sm = state.DBStateManager('testuser', 'osrtf')

    def test_gen_name(self):
        name = self.sm._gen_name('testuser', 'osrtf')
        self.assertEqual(name, 'testuser_osrtf')

    def test_state_access(self):
        now = datetime.datetime.utcnow()
        self.sm.set_state(now)
        result = self.sm.get_state()
        self.assertEqual(result, str(now))

    def test_metadata_access(self):
        metadata = {'foo': 'bar'}
        now = datetime.datetime.utcnow()
        self.sm.set_state(now)
        self.sm.set_metadata(metadata)
        result = self.sm.get_metadata()
        self.assertEqual(result, metadata)



# -*- coding: utf-8 -*-
# Copyright 2017 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Luka Peschke
#

import mock

from gnocchiclient import exceptions as gexc

from cloudkitty import storage
from cloudkitty.storage.hybrid.backends import gnocchi as hgnocchi
from cloudkitty import tests
from cloudkitty.tests import samples


class BaseHybridStorageTest(tests.TestCase):

    def setUp(self):
        super(BaseHybridStorageTest, self).setUp()
        self.conf.set_override('backend', 'hybrid', 'storage')
        hgnocchi.METRICS_CONF = samples.METRICS_CONF
        self.storage = storage.get_storage()
        with mock.patch.object(
                self.storage._hybrid_backend, 'init'):
            self.storage.init()


class PermissiveDict(object):
    """Allows to check a single key of a dict in an assertion.

    Example:
    >>> mydict = {'a': 'A', 'b': 'B'}
    >>> checker = PermissiveDict('A', key='a')
    >>> checker == mydict
    True
    """
    def __init__(self, value, key='name'):
        self.key = key
        self.value = value

    def __eq__(self, other):
        return self.value == other.get(self.key)


class HybridStorageTestGnocchi(BaseHybridStorageTest):

    def setUp(self):
        super(HybridStorageTestGnocchi, self).setUp()

    def tearDown(self):
        super(HybridStorageTestGnocchi, self).tearDown()

    def _init_storage(self, archive_policy=False, res_type=False):
        with mock.patch.object(self.storage._hybrid_backend._conn,
                               'archive_policy',
                               spec=['get', 'create']) as pol_mock:
            if not archive_policy:
                pol_mock.get.side_effect = gexc.ArchivePolicyNotFound
            else:
                pol_mock.create.side_effect = gexc.ArchivePolicyAlreadyExists
            with mock.patch.object(self.storage._hybrid_backend._conn,
                                   'resource_type',
                                   spec=['get', 'create']) as rtype_mock:
                if not res_type:
                    rtype_mock.get.side_effect = gexc.ResourceTypeNotFound
                else:
                    rtype_mock.create.side_effect \
                        = gexc.ResourceTypeAlreadyExists

                self.storage.init()
                rtype_data = self.storage._hybrid_backend._resource_type_data
                rtype_calls = list()
                for val in rtype_data.values():
                    rtype_calls.append(
                        mock.call(PermissiveDict(val['name'], key='name')))
                if res_type:
                    rtype_mock.create.assert_not_called()
                else:
                    rtype_mock.create.assert_has_calls(
                        rtype_calls, any_order=True)
            pol_mock.get.assert_called_once_with(
                self.storage._hybrid_backend._archive_policy_name)
            if archive_policy:
                pol_mock.create.assert_not_called()
            else:
                apolicy = {
                    'name': self.storage._hybrid_backend._archive_policy_name,
                    'back_window': 0,
                    'aggregation_methods':
                        ['std', 'count', 'min', 'max', 'sum', 'mean'],
                }
                apolicy['definition'] = \
                    self.storage._hybrid_backend._archive_policy_definition
                pol_mock.create.assert_called_once_with(apolicy)

    def test_init_no_res_type_no_policy(self):
        self._init_storage()

    def test_init_with_res_type_no_policy(self):
        self._init_storage(res_type=True)

    def test_init_no_res_type_with_policy(self):
        self._init_storage(archive_policy=True)

    def test_init_with_res_type_with_policy(self):
        self._init_storage(res_type=True, archive_policy=True)



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
from oslo_config import cfg
import pecan
from pecan import rest
from wsme import types as wtypes
import wsmeext.pecan as wsme_pecan

from cloudkitty.api.v1 import controllers as v1_api

CONF = cfg.CONF
CONF.import_opt('port', 'cloudkitty.api.app', 'api')


class APILink(wtypes.Base):
    """API link description.

    """

    type = wtypes.text
    """Type of link."""

    rel = wtypes.text
    """Relationship with this link."""

    href = wtypes.text
    """URL of the link."""

    @classmethod
    def sample(cls):
        version = 'v1'
        sample = cls(
            rel='self',
            type='text/html',
            href='http://127.0.0.1:{port}/{id}'.format(
                port=CONF.api.port,
                id=version))
        return sample


class APIMediaType(wtypes.Base):
    """Media type description.

    """

    base = wtypes.text
    """Base type of this media type."""

    type = wtypes.text
    """Type of this media type."""

    @classmethod
    def sample(cls):
        sample = cls(
            base='application/json',
            type='application/vnd.openstack.cloudkitty-v1+json')
        return sample


VERSION_STATUS = wtypes.Enum(wtypes.text, 'EXPERIMENTAL', 'STABLE')


class APIVersion(wtypes.Base):
    """API Version description.

    """

    id = wtypes.text
    """ID of the version."""

    status = VERSION_STATUS
    """Status of the version."""

    updated = wtypes.text
    "Last update in iso8601 format."

    links = [APILink]
    """List of links to API resources."""

    media_types = [APIMediaType]
    """Types accepted by this API."""

    @classmethod
    def sample(cls):
        version = 'v1'
        updated = '2014-08-11T16:00:00Z'
        links = [APILink.sample()]
        media_types = [APIMediaType.sample()]
        sample = cls(id=version,
                     status='STABLE',
                     updated=updated,
                     links=links,
                     media_types=media_types)
        return sample


class RootController(rest.RestController):
    """Root REST Controller exposing versions of the API.

    """

    v1 = v1_api.V1Controller()

    @wsme_pecan.wsexpose([APIVersion])
    def index(self):
        """Return the version list

        """
        # TODO(sheeprine): Maybe we should store all the API version
        # informations in every API modules
        ver1 = APIVersion(
            id='v1',
            status='EXPERIMENTAL',
            updated='2015-03-09T16:00:00Z',
            links=[
                APILink(
                    rel='self',
                    href='{scheme}://{host}/v1'.format(
                        scheme=pecan.request.scheme,
                        host=pecan.request.host,
                    )
                )
            ],
            media_types=[]
        )

        versions = []
        versions.append(ver1)

        return versions



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Add per tenant hashmap support

Revision ID: 4da82e1c11c8
Revises: c88a06b1cfce
Create Date: 2016-05-31 12:27:30.821497

"""

# revision identifiers, used by Alembic.
revision = '4da82e1c11c8'
down_revision = 'c88a06b1cfce'

from alembic import op
import sqlalchemy as sa

CONSTRAINT_MAP = {
    'hashmap_mappings': {
        u'uniq_field_mapping': (
            ['value', 'field_id', 'tenant_id'],
            ['value', 'field_id']),
        u'uniq_service_mapping': (
            ['value', 'service_id', 'tenant_id'],
            ['value', 'service_id'])},
    'hashmap_thresholds': {
        u'uniq_field_threshold': (
            ['level', 'field_id', 'tenant_id'],
            ['level', 'field_id']),
        u'uniq_service_threshold': (
            ['level', 'service_id', 'tenant_id'],
            ['level', 'service_id'])}}


def get_reflect(table):
    reflect_args = [
        sa.Column(
            'service_id',
            sa.Integer,
            sa.ForeignKey(
                'hashmap_services.id',
                ondelete='CASCADE',
                name='fk_{}_service_id_hashmap_services'.format(table)),
            nullable=True),
        sa.Column(
            'field_id',
            sa.Integer,
            sa.ForeignKey(
                'hashmap_fields.id',
                ondelete='CASCADE',
                name='fk_{}_field_id_hashmap_fields'.format(table)),
            nullable=True),
        sa.Column(
            'group_id',
            sa.Integer,
            sa.ForeignKey(
                'hashmap_groups.id',
                ondelete='SET NULL',
                name='fk_{}_group_id_hashmap_groups'.format(table)),
            nullable=True),
        sa.Column(
            'map_type',
            sa.Enum(
                'flat',
                'rate',
                name='enum_{}map_type'.format(
                    'hash' if table == 'hashmap_thresholds' else '')),
            nullable=False)]
    return reflect_args


def upgrade():
    for table in ('hashmap_mappings', 'hashmap_thresholds'):
        with op.batch_alter_table(
            table,
            reflect_args=get_reflect(table)
        ) as batch_op:
            batch_op.add_column(
                sa.Column(
                    'tenant_id',
                    sa.String(length=36),
                    nullable=True))
            for name, columns in CONSTRAINT_MAP[table].items():
                batch_op.drop_constraint(name, type_='unique')
                batch_op.create_unique_constraint(name, columns[0])



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import decimal

from oslo_config import cfg
from wsme import types as wtypes

from cloudkitty.api.v1 import types as cktypes
from cloudkitty import utils as ck_utils

CONF = cfg.CONF

METRICS_CONF = ck_utils.get_metrics_conf(CONF.collect.metrics_conf)

CLOUDKITTY_SERVICES = wtypes.Enum(wtypes.text,
                                  *METRICS_CONF['services'])


class CloudkittyResource(wtypes.Base):
    """Type describing a resource in CloudKitty.

    """

    service = CLOUDKITTY_SERVICES
    """Name of the service."""

    # FIXME(sheeprine): values should be dynamic
    # Testing with ironic dynamic type
    desc = {wtypes.text: cktypes.MultiType(wtypes.text, int, float,
                                           dict, decimal.Decimal)}
    """Description of the resources parameters."""

    volume = decimal.Decimal
    """Volume of resources."""

    def to_json(self):
        res_dict = {}
        res_dict[self.service] = [{'desc': self.desc,
                                   'vol': {'qty': str(self.volume),
                                           'unit': 'undef'}
                                   }]
        return res_dict

    @classmethod
    def sample(cls):
        sample = cls(service='compute',
                     desc={
                         'image_id': 'a41fba37-2429-4f15-aa00-b5bc4bf557bf'
                     },
                     volume=decimal.Decimal(1))
        return sample


class CloudkittyResourceCollection(wtypes.Base):
    """A list of CloudKittyResources."""

    resources = [CloudkittyResource]


class CloudkittyModule(wtypes.Base):
    """A rating extension summary

    """

    module_id = wtypes.wsattr(wtypes.text, mandatory=True)
    """Name of the extension."""

    description = wtypes.wsattr(wtypes.text, mandatory=False)
    """Short description of the extension."""

    enabled = wtypes.wsattr(bool)
    """Extension status."""

    hot_config = wtypes.wsattr(bool, default=False, name='hot-config')
    """On-the-fly configuration support."""

    priority = wtypes.wsattr(int)
    """Priority of the extension."""

    @classmethod
    def sample(cls):
        sample = cls(name='example',
                     description='Sample extension.',
                     enabled=True,
                     hot_config=False,
                     priority=2)
        return sample


class CloudkittyModuleCollection(wtypes.Base):
    """A list of rating extensions."""

    modules = [CloudkittyModule]



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
from logging import config as log_config

from alembic import context

from cloudkitty import db

config = context.config
log_config.fileConfig(config.config_file_name)


def run_migrations_online(target_metadata, version_table):
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    :param target_metadata: Model's metadata used for autogenerate support.
    :param version_table: Override the default version table for alembic.
    """
    engine = db.get_engine()
    with engine.connect() as connection:
        context.configure(connection=connection,
                          target_metadata=target_metadata,
                          version_table=version_table)
        with context.begin_transaction():
            context.run_migrations()



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Added threshold support.

Revision ID: 4fa888fd7eda
Revises: 3dd7e13527f3
Create Date: 2015-05-05 14:39:24.562388

"""

# revision identifiers, used by Alembic.
revision = '4fa888fd7eda'
down_revision = '3dd7e13527f3'

from alembic import op
import sqlalchemy as sa


def upgrade():
    # NOTE(sheeprine): Hack to let the migrations pass for postgresql
    dialect = op.get_context().dialect.name
    if dialect == 'postgresql':
        constraints = ['uniq_field_threshold', 'uniq_service_threshold']
    else:
        constraints = ['uniq_field_mapping', 'uniq_service_mapping']
    op.create_table(
        'hashmap_thresholds',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('threshold_id', sa.String(length=36), nullable=False),
        sa.Column('level', sa.Numeric(precision=20, scale=8), nullable=True),
        sa.Column('cost', sa.Numeric(precision=20, scale=8), nullable=False),
        sa.Column(
            'map_type',
            sa.Enum('flat', 'rate', name='enum_map_type'),
            nullable=False),
        sa.Column('service_id', sa.Integer(), nullable=True),
        sa.Column('field_id', sa.Integer(), nullable=True),
        sa.Column('group_id', sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ['field_id'],
            ['hashmap_fields.id'],
            ondelete='CASCADE'),
        sa.ForeignKeyConstraint(
            ['group_id'],
            ['hashmap_groups.id'],
            ondelete='SET NULL'),
        sa.ForeignKeyConstraint(
            ['service_id'],
            ['hashmap_services.id'],
            ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('threshold_id'),
        sa.UniqueConstraint('level', 'field_id', name=constraints[0]),
        sa.UniqueConstraint('level', 'service_id', name=constraints[1]),
        mysql_charset='utf8',
        mysql_engine='InnoDB')



# -*- coding: utf-8 -*-
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
from wsme import types as wtypes
import wsmeext.pecan as wsme_pecan

from cloudkitty import rating
from cloudkitty.rating.hash.controllers import field as field_api
from cloudkitty.rating.hash.controllers import group as group_api
from cloudkitty.rating.hash.controllers import mapping as mapping_api
from cloudkitty.rating.hash.controllers import service as service_api
from cloudkitty.rating.hash.controllers import threshold as threshold_api
from cloudkitty.rating.hash.datamodels import mapping as mapping_models


class HashMapConfigController(rating.RatingRestControllerBase):
    """Controller exposing all management sub controllers.

    """

    _custom_actions = {
        'types': ['GET']
    }

    services = service_api.HashMapServicesController()
    fields = field_api.HashMapFieldsController()
    groups = group_api.HashMapGroupsController()
    mappings = mapping_api.HashMapMappingsController()
    thresholds = threshold_api.HashMapThresholdsController()

    @wsme_pecan.wsexpose([wtypes.text])
    def get_types(self):
        """Return the list of every mapping type available.

        """
        return mapping_models.MAP_TYPE.values



# -*- coding: utf-8 -*-
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import copy
import decimal

from cloudkitty import utils as ck_utils

TENANT = 'f266f30b11f246b589fd266f85eeec39'
INITIAL_TIMESTAMP = 1420070400
FIRST_PERIOD_BEGIN = INITIAL_TIMESTAMP
FIRST_PERIOD_BEGIN_ISO = ck_utils.ts2iso(FIRST_PERIOD_BEGIN)
FIRST_PERIOD_END = FIRST_PERIOD_BEGIN + 3600
FIRST_PERIOD_END_ISO = ck_utils.ts2iso(FIRST_PERIOD_END)
SECOND_PERIOD_BEGIN = FIRST_PERIOD_END
SECOND_PERIOD_BEGIN_ISO = ck_utils.ts2iso(SECOND_PERIOD_BEGIN)
SECOND_PERIOD_END = SECOND_PERIOD_BEGIN + 3600
SECOND_PERIOD_END_ISO = ck_utils.ts2iso(SECOND_PERIOD_END)

COMPUTE_METADATA = {
    'availability_zone': 'nova',
    'flavor': 'm1.nano',
    'image_id': 'f5600101-8fa2-4864-899e-ebcb7ed6b568',
    'instance_id': '26c084e1-b8f1-4cbc-a7ec-e8b356788a17',
    'resource_id': '1558f911-b55a-4fd2-9173-c8f1f23e5639',
    'memory': '64',
    'metadata': {
        'farm': 'prod'
    },
    'name': 'prod1',
    'project_id': 'f266f30b11f246b589fd266f85eeec39',
    'user_id': '55b3379b949243009ee96972fbf51ed1',
    'vcpus': '1'}

IMAGE_METADATA = {
    'checksum': '836c69cbcd1dc4f225daedbab6edc7c7',
    'resource_id': '7b5b73f2-9181-4307-a710-b1aa6472526d',
    'container_format': 'aki',
    'created_at': '2014-06-04T16:26:01',
    'deleted': 'False',
    'deleted_at': 'None',
    'disk_format': 'aki',
    'is_public': 'True',
    'min_disk': '0',
    'min_ram': '0',
    'name': 'cirros-0.3.2-x86_64-uec-kernel',
    'protected': 'False',
    'size': '4969360',
    'status': 'active',
    'updated_at': '2014-06-04T16:26:02'}

FIRST_PERIOD = {
    'begin': FIRST_PERIOD_BEGIN,
    'end': FIRST_PERIOD_END}

SECOND_PERIOD = {
    'begin': SECOND_PERIOD_BEGIN,
    'end': SECOND_PERIOD_END}

COLLECTED_DATA = [{
    'period': FIRST_PERIOD,
    'usage': {
        'compute': [{
            'desc': COMPUTE_METADATA,
            'vol': {
                'qty': decimal.Decimal(1.0),
                'unit': 'instance'}}],
        'image': [{
            'desc': IMAGE_METADATA,
            'vol': {
                'qty': decimal.Decimal(1.0),
                'unit': 'image'}}]
    }}, {
    'period': SECOND_PERIOD,
    'usage': {
        'compute': [{
            'desc': COMPUTE_METADATA,
            'vol': {
                'qty': decimal.Decimal(1.0),
                'unit': 'instance'}}]
    }}]

RATED_DATA = copy.deepcopy(COLLECTED_DATA)
RATED_DATA[0]['usage']['compute'][0]['rating'] = {
    'price': decimal.Decimal('0.42')}
RATED_DATA[0]['usage']['image'][0]['rating'] = {
    'price': decimal.Decimal('0.1337')}
RATED_DATA[1]['usage']['compute'][0]['rating'] = {
    'price': decimal.Decimal('0.42')}


def split_storage_data(raw_data):
    final_data = []
    for frame in raw_data:
        frame['period']['begin'] = ck_utils.ts2iso(frame['period']['begin'])
        frame['period']['end'] = ck_utils.ts2iso(frame['period']['end'])
        usage_buffer = frame.pop('usage')
        # Sort to have a consistent result as we are converting it to a list
        for service, data in sorted(usage_buffer.items()):
            new_frame = copy.deepcopy(frame)
            new_frame['usage'] = {service: data}
            new_frame['usage'][service][0]['tenant_id'] = TENANT
            final_data.append(new_frame)
    return final_data


# FIXME(sheeprine): storage is not using decimal for rates, we need to
# transition to decimal.
STORED_DATA = copy.deepcopy(COLLECTED_DATA)
STORED_DATA[0]['usage']['compute'][0]['rating'] = {
    'price': 0.42}
STORED_DATA[0]['usage']['image'][0]['rating'] = {
    'price': 0.1337}
STORED_DATA[1]['usage']['compute'][0]['rating'] = {
    'price': 0.42}

STORED_DATA = split_storage_data(STORED_DATA)

METRICS_CONF = {
    'collector': 'gnocchi',
    'name': 'OpenStack',
    'period': 3600,
    'services': [
        'compute',
        'volume',
        'network.bw.in',
        'network.bw.out',
        'network.floating',
        'image'
    ],
    'services_metrics': {
        'compute': [
            {'vcpus': 'max'},
            {'memory': 'max'},
            {'cpu': 'max'},
            {'disk.root.size': 'max'},
            {'disk.ephemeral.size': 'max'}
        ],
        'image': [
            {'image.size': 'max'},
            {'image.download': 'max'},
            {'image.serve': 'max'}
        ],
        'network.bw.in': [{'network.incoming.bytes': 'max'}],
        'network.bw.out': [{'network.outgoing.bytes': 'max'}],
        'network.floating': [{'ip.floating': 'max'}],
        'volume': [{'volume.size': 'max'}],
        'radosgw.usage': [{'radosgw.objects.size': 'max'}]},
    'services_objects': {
        'compute': 'instance',
        'image': 'image',
        'network.bw.in': 'instance_network_interface',
        'network.bw.out': 'instance_network_interface',
        'network.floating': 'network',
        'volume': 'volume',
        'radosgw.usage': 'ceph_account',
    },
    'metrics_units': {
        'compute': {1: {'unit': 'instance'}},
        'default_unit': {1: {'unit': 'unknown'}},
        'image': {'image.size': {'unit': 'MiB', 'factor': '1/1048576'}},
        'network.bw.in': {'network.incoming.bytes': {
            'unit': 'MB',
            'factor': '1/1000000'}},
        'network.bw.out': {'network.outgoing.bytes': {
            'unit': 'MB',
            'factor': '1/1000000'}},
        'network.floating': {1: {'unit': 'ip'}},
        'volume': {'volume.size': {'unit': 'GiB'}},
        'radosgw.usage': {'radosgw.objects.size': {
            'unit': 'GiB',
            'factor': '1/1073741824'}},
    },
    'wait_periods': 2,
    'window': 1800
}



# -*- coding: utf-8 -*-
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import copy

import mock
import sqlalchemy
import testscenarios

from cloudkitty import storage
from cloudkitty.storage.hybrid.backends import gnocchi as hgnocchi
from cloudkitty import tests
from cloudkitty.tests import samples
from cloudkitty import utils as ck_utils


class StorageTest(tests.TestCase):
    storage_scenarios = [
        ('sqlalchemy', dict(storage_backend='sqlalchemy')),
        ('hybrid', dict(storage_backend='hybrid'))]

    @classmethod
    def generate_scenarios(cls):
        cls.scenarios = testscenarios.multiply_scenarios(
            cls.scenarios,
            cls.storage_scenarios)

    @mock.patch('cloudkitty.storage.hybrid.backends.gnocchi.gclient')
    def setUp(self, gclient_mock):
        super(StorageTest, self).setUp()
        hgnocchi.METRICS_CONF = samples.METRICS_CONF
        self._tenant_id = samples.TENANT
        self._other_tenant_id = '8d3ae50089ea4142-9c6e1269db6a0b64'
        self.conf.set_override('backend', self.storage_backend, 'storage')
        self.storage = storage.get_storage()
        self.storage.init()

    def insert_data(self):
        working_data = copy.deepcopy(samples.RATED_DATA)
        self.storage.append(working_data, self._tenant_id)
        working_data = copy.deepcopy(samples.RATED_DATA)
        self.storage.append(working_data, self._other_tenant_id)
        self.storage.commit(self._tenant_id)
        self.storage.commit(self._other_tenant_id)

    def insert_different_data_two_tenants(self):
        working_data = copy.deepcopy(samples.RATED_DATA)
        del working_data[1]
        self.storage.append(working_data, self._tenant_id)
        working_data = copy.deepcopy(samples.RATED_DATA)
        del working_data[0]
        self.storage.append(working_data, self._other_tenant_id)
        self.storage.commit(self._tenant_id)
        self.storage.commit(self._other_tenant_id)

    # Filtering
    def test_filter_period(self):
        working_data = copy.deepcopy(samples.RATED_DATA)
        usage_start, data = self.storage._filter_period(working_data)
        self.assertEqual(samples.FIRST_PERIOD_BEGIN, usage_start)
        self.assertEqual(samples.RATED_DATA[0]['usage'], data)
        expected_remaining_data = [{
            "period": samples.SECOND_PERIOD,
            "usage": samples.RATED_DATA[1]['usage']}]
        self.assertEqual(expected_remaining_data, working_data)
        usage_start, data = self.storage._filter_period(working_data)
        self.assertEqual(samples.SECOND_PERIOD_BEGIN, usage_start)
        self.assertEqual(samples.RATED_DATA[1]['usage'], data)
        self.assertEqual([], working_data)

    # State
    def test_get_state_when_nothing_in_storage(self):
        state = self.storage.get_state()
        self.assertIsNone(state)

    def test_get_latest_global_state(self):
        self.insert_different_data_two_tenants()
        state = self.storage.get_state()
        self.assertEqual(samples.SECOND_PERIOD_BEGIN, state)

    def test_get_state_on_rated_tenant(self):
        self.insert_different_data_two_tenants()
        state = self.storage.get_state(self._tenant_id)
        self.assertEqual(samples.FIRST_PERIOD_BEGIN, state)
        state = self.storage.get_state(self._other_tenant_id)
        self.assertEqual(samples.SECOND_PERIOD_BEGIN, state)

    def test_get_state_on_no_data_frame(self):
        self.storage.nodata(
            samples.FIRST_PERIOD_BEGIN,
            samples.FIRST_PERIOD_END,
            self._tenant_id)
        self.storage.commit(self._tenant_id)
        state = self.storage.get_state(self._tenant_id)
        self.assertEqual(samples.FIRST_PERIOD_BEGIN, state)


class StorageDataframeTest(StorageTest):

    storage_scenarios = [
        ('sqlalchemy', dict(storage_backend='sqlalchemy'))]

    # Queries
    # Data
    def test_get_no_frame_when_nothing_in_storage(self):
        self.assertRaises(
            storage.NoTimeFrame,
            self.storage.get_time_frame,
            begin=samples.FIRST_PERIOD_BEGIN - 3600,
            end=samples.FIRST_PERIOD_BEGIN)

    def test_get_frame_filter_outside_data(self):
        self.insert_different_data_two_tenants()
        self.assertRaises(
            storage.NoTimeFrame,
            self.storage.get_time_frame,
            begin=samples.FIRST_PERIOD_BEGIN - 3600,
            end=samples.FIRST_PERIOD_BEGIN)

    def test_get_frame_without_filter_but_timestamp(self):
        self.insert_different_data_two_tenants()
        data = self.storage.get_time_frame(
            begin=samples.FIRST_PERIOD_BEGIN,
            end=samples.SECOND_PERIOD_END)
        self.assertEqual(3, len(data))

    def test_get_frame_on_one_period(self):
        self.insert_different_data_two_tenants()
        data = self.storage.get_time_frame(
            begin=samples.FIRST_PERIOD_BEGIN,
            end=samples.FIRST_PERIOD_END)
        self.assertEqual(2, len(data))

    def test_get_frame_on_one_period_and_one_tenant(self):
        self.insert_different_data_two_tenants()
        data = self.storage.get_time_frame(
            begin=samples.FIRST_PERIOD_BEGIN,
            end=samples.FIRST_PERIOD_END,
            tenant_id=self._tenant_id)
        self.assertEqual(2, len(data))

    def test_get_frame_on_one_period_and_one_tenant_outside_data(self):
        self.insert_different_data_two_tenants()
        self.assertRaises(
            storage.NoTimeFrame,
            self.storage.get_time_frame,
            begin=samples.FIRST_PERIOD_BEGIN,
            end=samples.FIRST_PERIOD_END,
            tenant_id=self._other_tenant_id)

    def test_get_frame_on_two_periods(self):
        self.insert_different_data_two_tenants()
        data = self.storage.get_time_frame(
            begin=samples.FIRST_PERIOD_BEGIN,
            end=samples.SECOND_PERIOD_END)
        self.assertEqual(3, len(data))


class StorageTotalTest(StorageTest):

    storage_scenarios = [
        ('sqlalchemy', dict(storage_backend='sqlalchemy'))]

    # Total
    def test_get_empty_total(self):
        begin = ck_utils.ts2dt(samples.FIRST_PERIOD_BEGIN - 3600)
        end = ck_utils.ts2dt(samples.FIRST_PERIOD_BEGIN)
        self.insert_data()
        total = self.storage.get_total(
            begin=begin,
            end=end)
        self.assertEqual(1, len(total))
        self.assertIsNone(total[0]["rate"])
        self.assertEqual(begin, total[0]["begin"])
        self.assertEqual(end, total[0]["end"])

    def test_get_total_without_filter_but_timestamp(self):
        begin = ck_utils.ts2dt(samples.FIRST_PERIOD_BEGIN)
        end = ck_utils.ts2dt(samples.SECOND_PERIOD_END)
        self.insert_data()
        total = self.storage.get_total(
            begin=begin,
            end=end)
        # FIXME(sheeprine): floating point error (transition to decimal)
        self.assertEqual(1, len(total))
        self.assertEqual(1.9473999999999998, total[0]["rate"])
        self.assertEqual(begin, total[0]["begin"])
        self.assertEqual(end, total[0]["end"])

    def test_get_total_filtering_on_one_period(self):
        begin = ck_utils.ts2dt(samples.FIRST_PERIOD_BEGIN)
        end = ck_utils.ts2dt(samples.FIRST_PERIOD_END)
        self.insert_data()
        total = self.storage.get_total(
            begin=begin,
            end=end)
        self.assertEqual(1, len(total))
        self.assertEqual(1.1074, total[0]["rate"])
        self.assertEqual(begin, total[0]["begin"])
        self.assertEqual(end, total[0]["end"])

    def test_get_total_filtering_on_one_period_and_one_tenant(self):
        begin = ck_utils.ts2dt(samples.FIRST_PERIOD_BEGIN)
        end = ck_utils.ts2dt(samples.FIRST_PERIOD_END)
        self.insert_data()
        total = self.storage.get_total(
            begin=begin,
            end=end,
            tenant_id=self._tenant_id)
        self.assertEqual(1, len(total))
        self.assertEqual(0.5537, total[0]["rate"])
        self.assertEqual(self._tenant_id, total[0]["tenant_id"])
        self.assertEqual(begin, total[0]["begin"])
        self.assertEqual(end, total[0]["end"])

    def test_get_total_filtering_on_service(self):
        begin = ck_utils.ts2dt(samples.FIRST_PERIOD_BEGIN)
        end = ck_utils.ts2dt(samples.FIRST_PERIOD_END)
        self.insert_data()
        total = self.storage.get_total(
            begin=begin,
            end=end,
            service='compute')
        self.assertEqual(1, len(total))
        self.assertEqual(0.84, total[0]["rate"])
        self.assertEqual('compute', total[0]["res_type"])
        self.assertEqual(begin, total[0]["begin"])
        self.assertEqual(end, total[0]["end"])

    def test_get_total_groupby_tenant(self):
        begin = ck_utils.ts2dt(samples.FIRST_PERIOD_BEGIN)
        end = ck_utils.ts2dt(samples.SECOND_PERIOD_END)
        self.insert_data()
        total = self.storage.get_total(
            begin=begin,
            end=end,
            groupby="tenant_id")
        self.assertEqual(2, len(total))
        self.assertEqual(0.9737, total[0]["rate"])
        self.assertEqual(self._other_tenant_id, total[0]["tenant_id"])
        self.assertEqual(begin, total[0]["begin"])
        self.assertEqual(end, total[0]["end"])
        self.assertEqual(0.9737, total[1]["rate"])
        self.assertEqual(self._tenant_id, total[1]["tenant_id"])
        self.assertEqual(begin, total[1]["begin"])
        self.assertEqual(end, total[1]["end"])

    def test_get_total_groupby_restype(self):
        begin = ck_utils.ts2dt(samples.FIRST_PERIOD_BEGIN)
        end = ck_utils.ts2dt(samples.SECOND_PERIOD_END)
        self.insert_data()
        total = self.storage.get_total(
            begin=begin,
            end=end,
            groupby="res_type")
        self.assertEqual(2, len(total))
        self.assertEqual(0.2674, total[0]["rate"])
        self.assertEqual('image', total[0]["res_type"])
        self.assertEqual(begin, total[0]["begin"])
        self.assertEqual(end, total[0]["end"])
        self.assertEqual(1.68, total[1]["rate"])
        self.assertEqual('compute', total[1]["res_type"])
        self.assertEqual(begin, total[1]["begin"])
        self.assertEqual(end, total[1]["end"])

    def test_get_total_groupby_tenant_and_restype(self):
        begin = ck_utils.ts2dt(samples.FIRST_PERIOD_BEGIN)
        end = ck_utils.ts2dt(samples.SECOND_PERIOD_END)
        self.insert_data()
        total = self.storage.get_total(
            begin=begin,
            end=end,
            groupby="tenant_id,res_type")
        self.assertEqual(4, len(total))
        self.assertEqual(0.1337, total[0]["rate"])
        self.assertEqual(self._other_tenant_id, total[0]["tenant_id"])
        self.assertEqual('image', total[0]["res_type"])
        self.assertEqual(begin, total[0]["begin"])
        self.assertEqual(end, total[0]["end"])
        self.assertEqual(0.1337, total[1]["rate"])
        self.assertEqual(self._tenant_id, total[1]["tenant_id"])
        self.assertEqual('image', total[1]["res_type"])
        self.assertEqual(begin, total[1]["begin"])
        self.assertEqual(end, total[1]["end"])
        self.assertEqual(0.84, total[2]["rate"])
        self.assertEqual(self._other_tenant_id, total[2]["tenant_id"])
        self.assertEqual('compute', total[2]["res_type"])
        self.assertEqual(begin, total[2]["begin"])
        self.assertEqual(end, total[2]["end"])
        self.assertEqual(0.84, total[3]["rate"])
        self.assertEqual(self._tenant_id, total[3]["tenant_id"])
        self.assertEqual('compute', total[3]["res_type"])
        self.assertEqual(begin, total[3]["begin"])
        self.assertEqual(end, total[3]["end"])


class StorageTenantTest(StorageTest):

    storage_scenarios = [
        ('sqlalchemy', dict(storage_backend='sqlalchemy'))]

    # Tenants
    def test_get_empty_tenant_with_nothing_in_storage(self):
        tenants = self.storage.get_tenants(
            begin=ck_utils.ts2dt(samples.FIRST_PERIOD_BEGIN),
            end=ck_utils.ts2dt(samples.SECOND_PERIOD_BEGIN))
        self.assertEqual([], tenants)

    def test_get_empty_tenant_list(self):
        self.insert_data()
        tenants = self.storage.get_tenants(
            begin=ck_utils.ts2dt(samples.FIRST_PERIOD_BEGIN - 3600),
            end=ck_utils.ts2dt(samples.FIRST_PERIOD_BEGIN))
        self.assertEqual([], tenants)

    def test_get_tenants_filtering_on_period(self):
        self.insert_different_data_two_tenants()
        tenants = self.storage.get_tenants(
            begin=ck_utils.ts2dt(samples.FIRST_PERIOD_BEGIN),
            end=ck_utils.ts2dt(samples.SECOND_PERIOD_END))
        self.assertListEqual(
            [self._tenant_id, self._other_tenant_id],
            tenants)
        tenants = self.storage.get_tenants(
            begin=ck_utils.ts2dt(samples.FIRST_PERIOD_BEGIN),
            end=ck_utils.ts2dt(samples.FIRST_PERIOD_END))
        self.assertListEqual(
            [self._tenant_id],
            tenants)
        tenants = self.storage.get_tenants(
            begin=ck_utils.ts2dt(samples.SECOND_PERIOD_BEGIN),
            end=ck_utils.ts2dt(samples.SECOND_PERIOD_END))
        self.assertListEqual(
            [self._other_tenant_id],
            tenants)


class StorageDataIntegrityTest(StorageTest):

    storage_scenarios = [
        ('sqlalchemy', dict(storage_backend='sqlalchemy'))]

    # Data integrity
    def test_has_data_flag_behaviour(self):
        self.assertNotIn(self._tenant_id, self.storage._has_data)
        self.storage.nodata(
            samples.FIRST_PERIOD_BEGIN,
            samples.FIRST_PERIOD_END,
            self._tenant_id)
        self.assertNotIn(self._tenant_id, self.storage._has_data)
        working_data = copy.deepcopy(samples.RATED_DATA)
        working_data = [working_data[1]]
        self.storage.append(working_data, self._tenant_id)
        self.assertTrue(self.storage._has_data[self._tenant_id])
        self.storage.commit(self._tenant_id)
        self.assertNotIn(self._tenant_id, self.storage._has_data)

    def test_notify_no_data(self):
        self.storage.nodata(
            samples.FIRST_PERIOD_BEGIN,
            samples.FIRST_PERIOD_END,
            self._tenant_id)
        working_data = copy.deepcopy(samples.RATED_DATA)
        working_data = [working_data[1]]
        self.storage.append(working_data, self._tenant_id)
        kwargs = {
            'begin': samples.FIRST_PERIOD_BEGIN,
            'end': samples.FIRST_PERIOD_END,
            'tenant_id': self._tenant_id}
        self.assertRaises(
            storage.NoTimeFrame,
            self.storage.get_time_frame,
            **kwargs)
        kwargs['res_type'] = '_NO_DATA_'
        stored_data = self.storage.get_time_frame(**kwargs)
        self.assertEqual(1, len(stored_data))
        self.assertEqual(1, len(stored_data[0]['usage']))
        self.assertIn('_NO_DATA_', stored_data[0]['usage'])

    def test_send_nodata_between_data(self):
        working_data = copy.deepcopy(samples.RATED_DATA)
        for period in working_data:
            for service, data in sorted(period['usage'].items()):
                sub_data = [{
                    'period': period['period'],
                    'usage': {
                        service: data}}]
                self.storage.append(sub_data, self._tenant_id)
                if service == 'compute':
                    self.storage.nodata(
                        period['period']['begin'],
                        period['period']['end'],
                        self._tenant_id)
            self.storage.commit(self._tenant_id)
        self.assertRaises(
            storage.NoTimeFrame,
            self.storage.get_time_frame,
            begin=samples.FIRST_PERIOD_BEGIN,
            end=samples.SECOND_PERIOD_END,
            res_type='_NO_DATA_')

    def test_auto_commit_on_period_change(self):
        working_data = copy.deepcopy(samples.RATED_DATA)
        self.storage.append(working_data, self._tenant_id)
        stored_data = self.storage.get_time_frame(
            begin=samples.FIRST_PERIOD_BEGIN,
            end=samples.SECOND_PERIOD_END)
        self.assertEqual(2, len(stored_data))
        expected_data = copy.deepcopy(samples.STORED_DATA)
        # We only stored the first timeframe, the second one is waiting for a
        # commit or an append with the next timeframe.
        del expected_data[2]
        # NOTE(sheeprine): Quick and dirty sort (ensure result consistency,
        # order is not significant to the test result)
        if 'image' in stored_data[0]['usage']:
            stored_data[0]['usage'], stored_data[1]['usage'] = (
                stored_data[1]['usage'], stored_data[0]['usage'])
        self.assertEqual(
            expected_data,
            stored_data)

    def test_create_session_on_append(self):
        self.assertNotIn(self._tenant_id, self.storage._session)
        working_data = copy.deepcopy(samples.RATED_DATA)
        self.storage.append(working_data, self._tenant_id)
        self.assertIn(self._tenant_id, self.storage._session)
        self.assertIsInstance(
            self.storage._session[self._tenant_id],
            sqlalchemy.orm.session.Session)

    def test_delete_session_on_commit(self):
        working_data = copy.deepcopy(samples.RATED_DATA)
        self.storage.append(working_data, self._tenant_id)
        self.storage.commit(self._tenant_id)
        self.assertNotIn(self._tenant_id, self.storage._session)

    def test_update_period_on_append(self):
        self.assertNotIn(self._tenant_id, self.storage.usage_start)
        self.assertNotIn(self._tenant_id, self.storage.usage_start_dt)
        self.assertNotIn(self._tenant_id, self.storage.usage_end)
        self.assertNotIn(self._tenant_id, self.storage.usage_end_dt)
        working_data = copy.deepcopy(samples.RATED_DATA)
        self.storage.append([working_data[0]], self._tenant_id)
        self.assertEqual(
            self.storage.usage_start[self._tenant_id],
            samples.FIRST_PERIOD_BEGIN)
        self.assertEqual(
            self.storage.usage_start_dt[self._tenant_id],
            ck_utils.ts2dt(samples.FIRST_PERIOD_BEGIN))
        self.assertEqual(
            self.storage.usage_end[self._tenant_id],
            samples.FIRST_PERIOD_END)
        self.assertEqual(
            self.storage.usage_end_dt[self._tenant_id],
            ck_utils.ts2dt(samples.FIRST_PERIOD_END))
        self.storage.append([working_data[1]], self._tenant_id)
        self.assertEqual(
            self.storage.usage_start[self._tenant_id],
            samples.SECOND_PERIOD_BEGIN)
        self.assertEqual(
            self.storage.usage_start_dt[self._tenant_id],
            ck_utils.ts2dt(samples.SECOND_PERIOD_BEGIN))
        self.assertEqual(
            self.storage.usage_end[self._tenant_id],
            samples.SECOND_PERIOD_END)
        self.assertEqual(
            self.storage.usage_end_dt[self._tenant_id],
            ck_utils.ts2dt(samples.SECOND_PERIOD_END))

    def test_clear_period_info_on_commit(self):
        working_data = copy.deepcopy(samples.RATED_DATA)
        self.storage.append(working_data, self._tenant_id)
        self.storage.commit(self._tenant_id)
        self.assertNotIn(self._tenant_id, self.storage.usage_start)
        self.assertNotIn(self._tenant_id, self.storage.usage_start_dt)
        self.assertNotIn(self._tenant_id, self.storage.usage_end)
        self.assertNotIn(self._tenant_id, self.storage.usage_end_dt)


StorageTest.generate_scenarios()
StorageTotalTest.generate_scenarios()
StorageTenantTest.generate_scenarios()
StorageDataframeTest.generate_scenarios()
StorageDataIntegrityTest.generate_scenarios()



# -*- coding: utf-8 -*-
# !/usr/bin/env python
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import abc

from oslo_config import cfg
import six

fetchers_opts = [
    cfg.StrOpt('backend',
               default='keystone',
               help='Driver used to fetch tenant list.')
]

cfg.CONF.register_opts(fetchers_opts, 'tenant_fetcher')


@six.add_metaclass(abc.ABCMeta)
class BaseFetcher(object):
    """CloudKitty tenants fetcher.

    Provides Cloudkitty integration with a backend announcing ratable tenants.
    """

    @abc.abstractmethod
    def get_tenants(self):
        """Get a list of tenants to rate."""



# -*- coding: utf-8 -*-
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
from wsme import types as wtypes

from cloudkitty.api.v1 import types as ck_types


class Script(wtypes.Base):
    """Type describing a script.

    """

    script_id = wtypes.wsattr(ck_types.UuidType(), mandatory=False)
    """UUID of the script."""

    name = wtypes.wsattr(wtypes.text, mandatory=True)
    """Name of the script."""

    data = wtypes.wsattr(wtypes.text, mandatory=False)
    """Data of the script."""

    checksum = wtypes.wsattr(wtypes.text, mandatory=False, readonly=True)
    """Checksum of the script data."""

    @classmethod
    def sample(cls):
        sample = cls(script_id='bc05108d-f515-4984-8077-de319cbf35aa',
                     name='policy1',
                     data='return 0',
                     checksum='da39a3ee5e6b4b0d3255bfef95601890afd80709')
        return sample


class ScriptCollection(wtypes.Base):
    """Type describing a list of scripts.

    """

    scripts = [Script]
    """List of scripts."""

    @classmethod
    def sample(cls):
        sample = Script.sample()
        return cls(scripts=[sample])



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
from wsme import types as wtypes


class CollectorInfos(wtypes.Base):
    """Type describing a collector module.

    """

    name = wtypes.wsattr(wtypes.text, mandatory=False)
    """Name of the collector."""

    enabled = wtypes.wsattr(bool, mandatory=True)
    """State of the collector."""

    def to_json(self):
        res_dict = {'name': self.name,
                    'enabled': self.enabled}
        return res_dict

    @classmethod
    def sample(cls):
        sample = cls(name='ceilometer',
                     enabled=True)
        return sample


class ServiceToCollectorMapping(wtypes.Base):
    """Type describing a service to collector mapping.

    """

    service = wtypes.text
    """Name of the service."""

    collector = wtypes.text
    """Name of the collector."""

    def to_json(self):
        res_dict = {'service': self.service,
                    'collector': self.collector}
        return res_dict

    @classmethod
    def sample(cls):
        sample = cls(service='compute',
                     collector='ceilometer')
        return sample


class ServiceToCollectorMappingCollection(wtypes.Base):
    """Type describing a service to collector mapping collection.

    """

    mappings = [ServiceToCollectorMapping]
    """List of service to collector mappings."""

    def to_json(self):
        res_dict = {'mappings': self.mappings}
        return res_dict

    @classmethod
    def sample(cls):
        mapping = ServiceToCollectorMapping(service='compute',
                                            collector='ceilometer')
        sample = cls(mappings=[mapping])
        return sample



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Gauvain Pocentek
#
from keystonemiddleware import auth_token


class AuthTokenMiddleware(auth_token.AuthProtocol):
    """A subclass of keystone auth_token middleware.

    It avoids authentication on public routes.
    """

    def __init__(self, app, conf, public_api_routes=[]):
        self._public_routes = public_api_routes
        super(AuthTokenMiddleware, self).__init__(app, conf)

    def __call__(self, env, start_response):
        # Strip the / from the URL if we're not dealing with '/'
        path = env.get('PATH_INFO').rstrip('/') or '/'

        if path in self._public_routes:
            return self._app(env, start_response)

        return super(AuthTokenMiddleware, self).__call__(env, start_response)

    @classmethod
    def factory(cls, global_config, **local_conf):
        public_routes = local_conf.get('acl_public_routes', '')
        public_api_routes = [path.strip() for path in public_routes.split(',')]

        def _factory(app):
            return cls(app, global_config, public_api_routes=public_api_routes)
        return _factory



# -*- coding: utf-8 -*-
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import pecan
import six
import wsmeext.pecan as wsme_pecan

from cloudkitty.api.v1 import types as ck_types
from cloudkitty import rating
from cloudkitty.rating.hash.datamodels import field as field_models
from cloudkitty.rating.hash.db import api as db_api


class HashMapFieldsController(rating.RatingRestControllerBase):
    """Controller responsible of fields management.

    """

    @wsme_pecan.wsexpose(field_models.FieldCollection,
                         ck_types.UuidType(),
                         status_code=200)
    def get_all(self, service_id):
        """Get the field list.

        :param service_id: Service's UUID to filter on.
        :return: List of every fields.
        """
        hashmap = db_api.get_instance()
        field_list = []
        fields_uuid_list = hashmap.list_fields(service_id)
        for field_uuid in fields_uuid_list:
            field_db = hashmap.get_field(field_uuid)
            field_list.append(field_models.Field(
                **field_db.export_model()))
        res = field_models.FieldCollection(fields=field_list)
        return res

    @wsme_pecan.wsexpose(field_models.Field,
                         ck_types.UuidType(),
                         status_code=200)
    def get_one(self, field_id):
        """Return a field.

        :param field_id: UUID of the field to filter on.
        """
        hashmap = db_api.get_instance()
        try:
            field_db = hashmap.get_field(uuid=field_id)
            return field_models.Field(**field_db.export_model())
        except db_api.NoSuchField as e:
            pecan.abort(404, six.text_type(e))

    @wsme_pecan.wsexpose(field_models.Field,
                         body=field_models.Field,
                         status_code=201)
    def post(self, field_data):
        """Create a field.

        :param field_data: Informations about the field to create.
        """
        hashmap = db_api.get_instance()
        try:
            field_db = hashmap.create_field(
                field_data.service_id,
                field_data.name)
            pecan.response.location = pecan.request.path_url
            if pecan.response.location[-1] != '/':
                pecan.response.location += '/'
            pecan.response.location += field_db.field_id
            return field_models.Field(
                **field_db.export_model())
        except db_api.FieldAlreadyExists as e:
            pecan.abort(409, six.text_type(e))
        except db_api.ClientHashMapError as e:
            pecan.abort(400, six.text_type(e))

    @wsme_pecan.wsexpose(None,
                         ck_types.UuidType(),
                         status_code=204)
    def delete(self, field_id):
        """Delete the field and all the sub keys recursively.

        :param field_id: UUID of the field to delete.
        """
        hashmap = db_api.get_instance()
        try:
            hashmap.delete_field(uuid=field_id)
        except db_api.NoSuchField as e:
            pecan.abort(404, six.text_type(e))



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import os

from oslo_config import cfg
from oslo_log import log
from paste import deploy
import pecan

from cloudkitty.api import config as api_config
from cloudkitty.api import hooks
from cloudkitty import service
from cloudkitty import storage


LOG = log.getLogger(__name__)

auth_opts = [
    cfg.StrOpt('api_paste_config',
               default="api_paste.ini",
               help="Configuration file for WSGI definition of API."),
    cfg.StrOpt('auth_strategy',
               choices=['noauth', 'keystone'],
               default='keystone',
               help=("The strategy to use for auth. Supports noauth and "
                     "keystone")),
]

api_opts = [
    cfg.IPOpt('host_ip',
              default='0.0.0.0',
              help='The listen IP for the cloudkitty API server.'),
    cfg.PortOpt('port',
                default=8889,
                help='The port for the cloudkitty API server.'),
    cfg.BoolOpt('pecan_debug',
                default=False,
                help='Toggle Pecan Debug Middleware.'),
]

CONF = cfg.CONF
CONF.register_opts(auth_opts)
CONF.register_opts(api_opts, group='api')


def get_pecan_config():
    # Set up the pecan configuration
    filename = api_config.__file__.replace('.pyc', '.py')
    return pecan.configuration.conf_from_file(filename)


def setup_app(pecan_config=None, extra_hooks=None):

    app_conf = get_pecan_config()
    storage_backend = storage.get_storage()
    app_hooks = [
        hooks.RPCHook(),
        hooks.StorageHook(storage_backend),
        hooks.ContextHook(),
    ]

    app = pecan.make_app(
        app_conf.app.root,
        static_root=app_conf.app.static_root,
        template_path=app_conf.app.template_path,
        debug=CONF.api.pecan_debug,
        force_canonical=getattr(app_conf.app, 'force_canonical', True),
        hooks=app_hooks,
        guess_content_type_from_ext=False
    )

    return app


def load_app():
    cfg_file = None
    cfg_path = cfg.CONF.api_paste_config
    if not os.path.isabs(cfg_path):
        cfg_file = CONF.find_file(cfg_path)
    elif os.path.exists(cfg_path):
        cfg_file = cfg_path

    if not cfg_file:
        raise cfg.ConfigFilesNotFoundError([cfg.CONF.api_paste_config])
    LOG.info("Full WSGI config used: %s", cfg_file)
    appname = "cloudkitty+{}".format(cfg.CONF.auth_strategy)
    LOG.info("Cloudkitty api with '%s' auth type will be loaded.",
             cfg.CONF.auth_strategy)
    return deploy.loadapp("config:" + cfg_file, name=appname)


def build_wsgi_app(argv=None):
    service.prepare_service()
    return load_app()


def app_factory(global_config, **local_conf):
    return setup_app()



# -*- coding: utf-8 -*-
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import hashlib
import zlib

from oslo_db.sqlalchemy import models
import sqlalchemy
from sqlalchemy.ext import declarative
from sqlalchemy.ext import hybrid

Base = declarative.declarative_base()


class PyScriptsBase(models.ModelBase):
    __table_args__ = {'mysql_charset': "utf8",
                      'mysql_engine': "InnoDB"}
    fk_to_resolve = {}

    def save(self, session=None):
        from cloudkitty import db

        if session is None:
            session = db.get_session()

        super(PyScriptsBase, self).save(session=session)

    def as_dict(self):
        d = {}
        for c in self.__table__.columns:
            if c.name == 'id':
                continue
            d[c.name] = self[c.name]
        return d

    def _recursive_resolve(self, path):
        obj = self
        for attr in path.split('.'):
            if hasattr(obj, attr):
                obj = getattr(obj, attr)
            else:
                return None
        return obj

    def export_model(self):
        res = self.as_dict()
        for fk, mapping in self.fk_to_resolve.items():
            res[fk] = self._recursive_resolve(mapping)
        return res


class PyScriptsScript(Base, PyScriptsBase):
    """A PyScripts entry.

    """
    __tablename__ = 'pyscripts_scripts'

    id = sqlalchemy.Column(sqlalchemy.Integer,
                           primary_key=True)
    script_id = sqlalchemy.Column(sqlalchemy.String(36),
                                  nullable=False,
                                  unique=True)
    name = sqlalchemy.Column(
        sqlalchemy.String(255),
        nullable=False,
        unique=True)
    _data = sqlalchemy.Column('data',
                              sqlalchemy.LargeBinary(),
                              nullable=False)
    _checksum = sqlalchemy.Column('checksum',
                                  sqlalchemy.String(40),
                                  nullable=False)

    @hybrid.hybrid_property
    def data(self):
        udata = zlib.decompress(self._data)
        return udata

    @data.setter
    def data(self, value):
        sha_check = hashlib.sha1()
        sha_check.update(value)
        self._checksum = sha_check.hexdigest()
        self._data = zlib.compress(value)

    @hybrid.hybrid_property
    def checksum(self):
        return self._checksum

    def __repr__(self):
        return ('<PyScripts Script[{uuid}]: '
                'name={name}>').format(
                    uuid=self.script_id,
                    name=self.name)



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import mock
from oslo_messaging import conffixture
from stevedore import extension

from cloudkitty import orchestrator
from cloudkitty import tests


class FakeKeystoneClient(object):

    class FakeTenants(object):
        def list(self):
            return ['f266f30b11f246b589fd266f85eeec39',
                    '4dfb25b0947c4f5481daf7b948c14187']

    tenants = FakeTenants()


class OrchestratorTest(tests.TestCase):
    def setUp(self):
        super(OrchestratorTest, self).setUp()
        messaging_conf = self.useFixture(conffixture.ConfFixture(self.conf))
        messaging_conf.transport_driver = 'fake'
        self.conf.set_override('backend', 'keystone', 'tenant_fetcher')
        self.conf.import_group('keystone_fetcher',
                               'cloudkitty.tenant_fetcher.keystone')

    def setup_fake_modules(self):
        fake_module1 = tests.FakeRatingModule()
        fake_module1.module_name = 'fake1'
        fake_module1.set_priority(3)
        fake_module2 = tests.FakeRatingModule()
        fake_module2.module_name = 'fake2'
        fake_module2.set_priority(1)
        fake_module3 = tests.FakeRatingModule()
        fake_module3.module_name = 'fake3'
        fake_module3.set_priority(2)
        fake_extensions = [
            extension.Extension(
                'fake1',
                'cloudkitty.tests.FakeRatingModule1',
                None,
                fake_module1),
            extension.Extension(
                'fake2',
                'cloudkitty.tests.FakeRatingModule2',
                None,
                fake_module2),
            extension.Extension(
                'fake3',
                'cloudkitty.tests.FakeRatingModule3',
                None,
                fake_module3)]
        return fake_extensions

    def test_processors_ordering_in_workers(self):
        fake_extensions = self.setup_fake_modules()
        ck_ext_mgr = 'cloudkitty.extension_manager.EnabledExtensionManager'
        with mock.patch(ck_ext_mgr) as stevemock:
            fake_mgr = extension.ExtensionManager.make_test_instance(
                fake_extensions,
                'cloudkitty.rating.processors')
            stevemock.return_value = fake_mgr
            worker = orchestrator.BaseWorker()
            stevemock.assert_called_once_with(
                'cloudkitty.rating.processors',
                invoke_kwds={'tenant_id': None})
            self.assertEqual('fake1', worker._processors[0].name)
            self.assertEqual(3, worker._processors[0].obj.priority)
            self.assertEqual('fake3', worker._processors[1].name)
            self.assertEqual(2, worker._processors[1].obj.priority)
            self.assertEqual('fake2', worker._processors[2].name)
            self.assertEqual(1, worker._processors[2].obj.priority)



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
from oslo_config import cfg
from wsme import types as wtypes

from cloudkitty import utils as ck_utils

CONF = cfg.CONF

METRICS_CONF = ck_utils.get_metrics_conf(CONF.collect.metrics_conf)

CLOUDKITTY_SERVICES = wtypes.Enum(wtypes.text,
                                  *METRICS_CONF['services'])


class CloudkittyServiceInfo(wtypes.Base):
    """Type describing a service info in CloudKitty.

    """

    service_id = CLOUDKITTY_SERVICES
    """Name of the service."""

    metadata = [wtypes.text]
    """List of service metadata"""

    unit = wtypes.text
    """service unit"""

    def to_json(self):
        res_dict = {}
        res_dict[self.service_id] = [{
            'metadata': self.metadata,
            'unit': self.unit
        }]
        return res_dict

    @classmethod
    def sample(cls):
        sample = cls(service_id='compute',
                     metadata=['resource_id', 'flavor', 'availability_zone'],
                     unit='instance')
        return sample


class CloudkittyServiceInfoCollection(wtypes.Base):
    """A list of CloudKittyServiceInfo."""

    services = [CloudkittyServiceInfo]

    @classmethod
    def sample(cls):
        sample = CloudkittyServiceInfo.sample()
        return cls(services=[sample])



# -*- coding: utf-8 -*-
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import copy
import decimal
import hashlib
import zlib

import mock
from oslo_utils import uuidutils
import six

from cloudkitty.rating import pyscripts
from cloudkitty.rating.pyscripts.db import api
from cloudkitty import tests

FAKE_UUID = '6c1b8a30-797f-4b7e-ad66-9879b79059fb'
CK_RESOURCES_DATA = [{
    "period": {
        "begin": "2014-10-01T00:00:00",
        "end": "2014-10-01T01:00:00"},
    "usage": {
        "compute": [
            {
                "desc": {
                    "availability_zone": "nova",
                    "flavor": "m1.nano",
                    "image_id": "f5600101-8fa2-4864-899e-ebcb7ed6b568",
                    "memory": "64",
                    "metadata": {
                        "farm": "prod"},
                    "name": "prod1",
                    "project_id": "f266f30b11f246b589fd266f85eeec39",
                    "user_id": "55b3379b949243009ee96972fbf51ed1",
                    "vcpus": "1"},
                "vol": {
                    "qty": 1,
                    "unit": "instance"}
            },
            {
                "desc": {
                    "availability_zone": "nova",
                    "flavor": "m1.tiny",
                    "image_id": "a41fba37-2429-4f15-aa00-b5bc4bf557bf",
                    "memory": "512",
                    "metadata": {
                        "farm": "dev"},
                    "name": "dev1",
                    "project_id": "f266f30b11f246b589fd266f85eeec39",
                    "user_id": "55b3379b949243009ee96972fbf51ed1",
                    "vcpus": "1"},
                "vol": {
                    "qty": 2,
                    "unit": "instance"}},
            {
                "desc": {
                    "availability_zone": "nova",
                    "flavor": "m1.nano",
                    "image_id": "a41fba37-2429-4f15-aa00-b5bc4bf557bf",
                    "memory": "64",
                    "metadata": {
                        "farm": "dev"},
                    "name": "dev2",
                    "project_id": "f266f30b11f246b589fd266f85eeec39",
                    "user_id": "55b3379b949243009ee96972fbf51ed1",
                    "vcpus": "1"},
                "vol": {
                    "qty": 1,
                    "unit": "instance"}}]}}]

TEST_CODE1 = 'a = 1'.encode('utf-8')
TEST_CODE1_CHECKSUM = hashlib.sha1(TEST_CODE1).hexdigest()
TEST_CODE2 = 'a = 0'.encode('utf-8')
TEST_CODE2_CHECKSUM = hashlib.sha1(TEST_CODE2).hexdigest()
TEST_CODE3 = 'if a == 1: raise Exception()'.encode('utf-8')
TEST_CODE3_CHECKSUM = hashlib.sha1(TEST_CODE3).hexdigest()

COMPLEX_POLICY1 = """
import decimal


for period in data:
    for service, resources in period['usage'].items():
        if service == 'compute':
            for resource in resources:
                if resource['desc'].get('flavor') == 'm1.nano':
                    resource['rating'] = {
                        'price': decimal.Decimal(1.0)}
""".encode('utf-8')


class PyScriptsRatingTest(tests.TestCase):
    def setUp(self):
        super(PyScriptsRatingTest, self).setUp()
        self._tenant_id = 'f266f30b11f246b589fd266f85eeec39'
        self._db_api = pyscripts.PyScripts.db_api
        self._db_api.get_migration().upgrade('head')
        self._pyscripts = pyscripts.PyScripts(self._tenant_id)

    # Scripts tests
    @mock.patch.object(uuidutils, 'generate_uuid',
                       return_value=FAKE_UUID)
    def test_create_script(self, patch_generate_uuid):
        self._db_api.create_script('policy1', TEST_CODE1)
        scripts = self._db_api.list_scripts()
        self.assertEqual([FAKE_UUID], scripts)
        patch_generate_uuid.assert_called_once_with()

    def test_create_duplicate_script(self):
        self._db_api.create_script('policy1', TEST_CODE1)
        self.assertRaises(api.ScriptAlreadyExists,
                          self._db_api.create_script,
                          'policy1',
                          TEST_CODE1)

    def test_get_script_by_uuid(self):
        expected = self._db_api.create_script('policy1', TEST_CODE1)
        actual = self._db_api.get_script(uuid=expected.script_id)
        self.assertEqual(expected.data, actual.data)

    def test_get_script_by_name(self):
        expected = self._db_api.create_script('policy1', TEST_CODE1)
        actual = self._db_api.get_script(expected.name)
        self.assertEqual(expected.data, actual.data)

    def test_get_script_without_parameters(self):
        self._db_api.create_script('policy1', TEST_CODE1)
        self.assertRaises(
            ValueError,
            self._db_api.get_script)

    def test_delete_script_by_name(self):
        self._db_api.create_script('policy1', TEST_CODE1)
        self._db_api.delete_script('policy1')
        scripts = self._db_api.list_scripts()
        self.assertEqual([], scripts)

    def test_delete_script_by_uuid(self):
        script_db = self._db_api.create_script('policy1', TEST_CODE1)
        self._db_api.delete_script(uuid=script_db.script_id)
        scripts = self._db_api.list_scripts()
        self.assertEqual([], scripts)

    def test_delete_script_without_parameters(self):
        self._db_api.create_script('policy1', TEST_CODE1)
        self.assertRaises(
            ValueError,
            self._db_api.delete_script)

    def test_delete_unknown_script_by_name(self):
        self.assertRaises(api.NoSuchScript,
                          self._db_api.delete_script,
                          'dummy')

    def test_delete_unknown_script_by_uuid(self):
        self.assertRaises(
            api.NoSuchScript,
            self._db_api.delete_script,
            uuid='6e8de9fc-ee17-4b60-b81a-c9320e994e76')

    def test_update_script(self):
        script_db = self._db_api.create_script('policy1', TEST_CODE1)
        self._db_api.update_script(script_db.script_id, data=TEST_CODE2)
        actual = self._db_api.get_script(uuid=script_db.script_id)
        self.assertEqual(TEST_CODE2, actual.data)

    def test_update_script_uuid_disabled(self):
        expected = self._db_api.create_script('policy1', TEST_CODE1)
        self._db_api.update_script(expected.script_id,
                                   data=TEST_CODE2,
                                   script_id='42')
        actual = self._db_api.get_script(uuid=expected.script_id)
        self.assertEqual(expected.script_id, actual.script_id)

    def test_update_script_unknown_attribute(self):
        expected = self._db_api.create_script('policy1', TEST_CODE1)
        self.assertRaises(
            ValueError,
            self._db_api.update_script,
            expected.script_id,
            nonexistent=1)

    def test_empty_script_update(self):
        expected = self._db_api.create_script('policy1', TEST_CODE1)
        self.assertRaises(
            ValueError,
            self._db_api.update_script,
            expected.script_id)

    # Storage tests
    def test_compressed_data(self):
        data = TEST_CODE1
        self._db_api.create_script('policy1', data)
        script = self._db_api.get_script('policy1')
        expected = zlib.compress(data)
        self.assertEqual(expected, script._data)

    def test_on_the_fly_decompression(self):
        data = TEST_CODE1
        self._db_api.create_script('policy1', data)
        script = self._db_api.get_script('policy1')
        self.assertEqual(data, script.data)

    def test_script_repr(self):
        script_db = self._db_api.create_script('policy1', TEST_CODE1)
        self.assertEqual(
            '<PyScripts Script[{uuid}]: name={name}>'.format(
                uuid=script_db.script_id,
                name=script_db.name),
            six.text_type(script_db))

    # Checksum tests
    def test_validate_checksum(self):
        self._db_api.create_script('policy1', TEST_CODE1)
        script = self._db_api.get_script('policy1')
        self.assertEqual(TEST_CODE1_CHECKSUM, script.checksum)

    def test_read_only_checksum(self):
        self._db_api.create_script('policy1', TEST_CODE1)
        script = self._db_api.get_script('policy1')
        self.assertRaises(
            AttributeError,
            setattr,
            script,
            'checksum',
            'da39a3ee5e6b4b0d3255bfef95601890afd80709')

    def test_update_checksum(self):
        self._db_api.create_script('policy1', TEST_CODE1)
        script = self._db_api.get_script('policy1')
        script = self._db_api.update_script(script.script_id, data=TEST_CODE2)
        self.assertEqual(TEST_CODE2_CHECKSUM, script.checksum)

    # Code exec tests
    def test_load_scripts(self):
        policy1_db = self._db_api.create_script('policy1', TEST_CODE1)
        policy2_db = self._db_api.create_script('policy2', TEST_CODE2)
        self._pyscripts.load_scripts_in_memory()
        self.assertIn(policy1_db.script_id, self._pyscripts._scripts)
        self.assertIn(policy2_db.script_id, self._pyscripts._scripts)

    def test_purge_old_scripts(self):
        policy1_db = self._db_api.create_script('policy1', TEST_CODE1)
        policy2_db = self._db_api.create_script('policy2', TEST_CODE2)
        self._pyscripts.reload_config()
        self.assertIn(policy1_db.script_id, self._pyscripts._scripts)
        self.assertIn(policy2_db.script_id, self._pyscripts._scripts)
        self._db_api.delete_script(uuid=policy1_db.script_id)
        self._pyscripts.reload_config()
        self.assertNotIn(policy1_db.script_id, self._pyscripts._scripts)
        self.assertIn(policy2_db.script_id, self._pyscripts._scripts)

    @mock.patch.object(uuidutils, 'generate_uuid',
                       return_value=FAKE_UUID)
    def test_valid_script_data_loaded(self, patch_generate_uuid):
        self._db_api.create_script('policy1', TEST_CODE1)
        self._pyscripts.load_scripts_in_memory()
        expected = {
            FAKE_UUID: {
                'code': compile(
                    TEST_CODE1,
                    '<PyScripts: {name}>'.format(name='policy1'),
                    'exec'),
                'checksum': TEST_CODE1_CHECKSUM,
                'name': 'policy1'
            }}
        self.assertEqual(expected, self._pyscripts._scripts)
        context = {'a': 0}
        exec(self._pyscripts._scripts[FAKE_UUID]['code'], context)
        self.assertEqual(1, context['a'])

    def test_update_script_on_checksum_change(self):
        policy_db = self._db_api.create_script('policy1', TEST_CODE1)
        self._pyscripts.reload_config()
        self._db_api.update_script(policy_db.script_id, data=TEST_CODE2)
        self._pyscripts.reload_config()
        self.assertEqual(
            TEST_CODE2_CHECKSUM,
            self._pyscripts._scripts[policy_db.script_id]['checksum'])

    def test_exec_code_isolation(self):
        self._db_api.create_script('policy1', TEST_CODE1)
        self._db_api.create_script('policy2', TEST_CODE3)
        self._pyscripts.reload_config()
        self.assertRaises(NameError, self._pyscripts.process, {})

    # Processing
    def test_process_rating(self):
        self._db_api.create_script('policy1', COMPLEX_POLICY1)
        self._pyscripts.reload_config()
        actual_data = copy.deepcopy(CK_RESOURCES_DATA)
        expected_data = copy.deepcopy(CK_RESOURCES_DATA)
        compute_list = expected_data[0]['usage']['compute']
        compute_list[0]['rating'] = {'price': decimal.Decimal('1')}
        compute_list[2]['rating'] = {'price': decimal.Decimal('1')}
        self._pyscripts.process(actual_data)
        self.assertEqual(expected_data, actual_data)



# -*- coding: utf-8 -*-
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import pecan
import six
import wsmeext.pecan as wsme_pecan

from cloudkitty.api.v1 import types as ck_types
from cloudkitty import rating
from cloudkitty.rating.hash.datamodels import group as group_models
from cloudkitty.rating.hash.datamodels import mapping as mapping_models
from cloudkitty.rating.hash.datamodels import threshold as threshold_models
from cloudkitty.rating.hash.db import api as db_api


class HashMapGroupsController(rating.RatingRestControllerBase):
    """Controller responsible of groups management.

    """
    _custom_actions = {
        'mappings': ['GET'],
        'thresholds': ['GET']}

    @wsme_pecan.wsexpose(mapping_models.MappingCollection,
                         ck_types.UuidType())
    def mappings(self, group_id):
        """Get the mappings attached to the group.

        :param group_id: UUID of the group to filter on.
        """
        hashmap = db_api.get_instance()
        mapping_list = []
        mappings_uuid_list = hashmap.list_mappings(group_uuid=group_id)
        for mapping_uuid in mappings_uuid_list:
            mapping_db = hashmap.get_mapping(uuid=mapping_uuid)
            mapping_list.append(mapping_models.Mapping(
                **mapping_db.export_model()))
        res = mapping_models.MappingCollection(mappings=mapping_list)
        return res

    @wsme_pecan.wsexpose(threshold_models.ThresholdCollection,
                         ck_types.UuidType())
    def thresholds(self, group_id):
        """Get the thresholds attached to the group.

        :param group_id: UUID of the group to filter on.
        """
        hashmap = db_api.get_instance()
        threshold_list = []
        thresholds_uuid_list = hashmap.list_thresholds(group_uuid=group_id)
        for threshold_uuid in thresholds_uuid_list:
            threshold_db = hashmap.get_threshold(uuid=threshold_uuid)
            threshold_list.append(threshold_models.Threshold(
                **threshold_db.export_model()))
        res = threshold_models.ThresholdCollection(thresholds=threshold_list)
        return res

    @wsme_pecan.wsexpose(group_models.GroupCollection)
    def get_all(self):
        """Get the group list

        :return: List of every group.
        """
        hashmap = db_api.get_instance()
        group_list = []
        groups_uuid_list = hashmap.list_groups()
        for group_uuid in groups_uuid_list:
            group_db = hashmap.get_group(uuid=group_uuid)
            group_list.append(group_models.Group(
                **group_db.export_model()))
        res = group_models.GroupCollection(groups=group_list)
        return res

    @wsme_pecan.wsexpose(group_models.Group,
                         ck_types.UuidType())
    def get_one(self, group_id):
        """Return a group.

        :param group_id: UUID of the group to filter on.
        """
        hashmap = db_api.get_instance()
        try:
            group_db = hashmap.get_group(uuid=group_id)
            return group_models.Group(**group_db.export_model())
        except db_api.NoSuchGroup as e:
            pecan.abort(404, six.text_type(e))

    @wsme_pecan.wsexpose(group_models.Group,
                         body=group_models.Group,
                         status_code=201)
    def post(self, group_data):
        """Create a group.

        :param group_data: Informations about the group to create.
        """
        hashmap = db_api.get_instance()
        try:
            group_db = hashmap.create_group(group_data.name)
            pecan.response.location = pecan.request.path_url
            if pecan.response.location[-1] != '/':
                pecan.response.location += '/'
            pecan.response.location += group_db.group_id
            return group_models.Group(
                **group_db.export_model())
        except db_api.GroupAlreadyExists as e:
            pecan.abort(409, six.text_type(e))
        except db_api.ClientHashMapError as e:
            pecan.abort(400, six.text_type(e))

    @wsme_pecan.wsexpose(None,
                         ck_types.UuidType(),
                         bool,
                         status_code=204)
    def delete(self, group_id, recursive=False):
        """Delete a group.

        :param group_id: UUID of the group to delete.
        :param recursive: Delete mappings recursively.
        """
        hashmap = db_api.get_instance()
        try:
            hashmap.delete_group(uuid=group_id, recurse=recursive)
        except db_api.NoSuchGroup as e:
            pecan.abort(404, six.text_type(e))



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Clean hashmap fields constraints.

Revision ID: c88a06b1cfce
Revises: f8c799db4aa0
Create Date: 2016-05-19 18:06:43.315066

"""

# revision identifiers, used by Alembic.
revision = 'c88a06b1cfce'
down_revision = 'f8c799db4aa0'

from alembic import op
import sqlalchemy as sa


def upgrade():
    with op.batch_alter_table(
        'hashmap_fields',
        # NOTE(sheeprine): Forced reflection is needed because of SQLAlchemy's
        # SQLite backend limitation reflecting ON DELETE clauses.
        reflect_args=[
            sa.Column(
                'service_id',
                sa.Integer,
                sa.ForeignKey(
                    'hashmap_services.id',
                    ondelete='CASCADE',
                    name='fk_hashmap_fields_service_id_hashmap_services'),
                nullable=False)]) as batch_op:
        batch_op.drop_constraint(
            u'uniq_field',
            type_='unique')
        batch_op.create_unique_constraint(
            'uniq_field_per_service',
            ['service_id', 'name'])
        batch_op.drop_constraint(
            'uniq_map_service_field',
            type_='unique')



# -*- coding: utf-8 -*-
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import unittest

import mock
from oslo_utils import uuidutils

from cloudkitty.tenant_fetcher import keystone
from cloudkitty import tests


class FakeRole(object):
    def __init__(self, name, uuid=None):
        self.id = uuid if uuid else uuidutils.generate_uuid()
        self.name = name


class FakeTenant(object):
    def __init__(self, id):
        self.id = id


class FakeKeystoneClient(object):

    user_id = 'd89e3fee-2b92-4387-b564-63901d62e591'

    def __init__(self, **kwargs):
        pass

    class FakeTenants(object):
        @classmethod
        def list(cls):
            return [FakeTenant('f266f30b11f246b589fd266f85eeec39'),
                    FakeTenant('4dfb25b0947c4f5481daf7b948c14187')]

    class FakeRoles(object):
        roles_mapping = {
            'd89e3fee-2b92-4387-b564-63901d62e591': {
                'f266f30b11f246b589fd266f85eeec39': [FakeRole('rating'),
                                                     FakeRole('admin')],
                '4dfb25b0947c4f5481daf7b948c14187': [FakeRole('admin')]}}

        @classmethod
        def roles_for_user(cls, user_id, tenant, **kwargs):
            return cls.roles_mapping[user_id][tenant.id]

    roles = FakeRoles()
    tenants = FakeTenants()


def Client(**kwargs):
    return FakeKeystoneClient(**kwargs)


class KeystoneFetcherTest(tests.TestCase):
    def setUp(self):
        super(KeystoneFetcherTest, self).setUp()
        self.conf.set_override('backend', 'keystone', 'tenant_fetcher')
        self.conf.import_group('keystone_fetcher',
                               'cloudkitty.tenant_fetcher.keystone')

    @unittest.SkipTest
    def test_keystone_tenant_fetcher_filter_list(self):
        kclient = 'keystoneclient.client.Client'
        with mock.patch(kclient) as kclientmock:
            kclientmock.return_value = Client()
            fetcher = keystone.KeystoneFetcher()
            kclientmock.assert_called_once_with(
                auth_url='http://127.0.0.1:5000/v2.0',
                username='cloudkitty',
                password='cloudkitty',
                tenant_name='cloudkitty',
                region_name='RegionOne')
            tenants = fetcher.get_tenants()
            self.assertEqual(['f266f30b11f246b589fd266f85eeec39'], tenants)



# -*- coding: utf-8 -*-
# Copyright 2016 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import copy

from cloudkitty import tests
from cloudkitty.tests import transformers as t_transformers
from cloudkitty.transformer import ceilometer

CEIL_COMPUTE = {
    'availability_zone': 'nova',
    'display_name': 'test',
    'instance_type': 'm1.nano',
    'image.id': 'c8ae2e38-316d-11e6-b19a-dbee663ddaee',
    'memory_mb': 64,
    'user_metadata.meta1': 'test1',
    'vcpus': '1'}

TRANS_COMPUTE = {
    'instance_id': '2f58a438-3169-11e6-b36c-bfe1fa3241fe',
    'project_id': '4480c638-3169-11e6-91de-a3bd3a7d3afb',
    'user_id': '576808d8-3169-11e6-992b-5f931fc671df',
    'availability_zone': 'nova',
    'metadata': {
        'meta1': 'test1'},
    'name': 'test',
    'flavor': 'm1.nano',
    'image_id': 'c8ae2e38-316d-11e6-b19a-dbee663ddaee',
    'memory': 64,
    'vcpus': '1'}

CEIL_VOLUME = {
    'availability_zone': 'az1',
    'volume_id': '17992d58-316f-11e6-9528-1379eed8ebe4',
    'display_name': 'vol1',
    'size': 10}

TRANS_VOLUME = {
    'volume_id': '17992d58-316f-11e6-9528-1379eed8ebe4',
    'project_id': '4480c638-3169-11e6-91de-a3bd3a7d3afb',
    'user_id': '576808d8-3169-11e6-992b-5f931fc671df',
    'availability_zone': 'az1',
    'name': 'vol1',
    'size': 10}

CEIL_IMAGE = {
    'status': 'active',
    'name': 'Cirros',
    'deleted': 'False',
    'disk_format': 'ami',
    'id': 'c4a0d12e-88ff-43e1-b182-f95dfe75e40c',
    'protected': 'False',
    'container_format': 'ami',
    'is_public': 'False',
    'size': '25165824'}

TRANS_IMAGE = {
    'image_id': '2f58a438-3169-11e6-b36c-bfe1fa3241fe',
    'project_id': '4480c638-3169-11e6-91de-a3bd3a7d3afb',
    'user_id': '576808d8-3169-11e6-992b-5f931fc671df',
    'container_format': 'ami',
    'deleted': 'False',
    'disk_format': 'ami',
    'is_public': 'False',
    'name': 'Cirros',
    'protected': 'False',
    'size': '25165824',
    'status': 'active'}

CEIL_NETWORK_TAP = {
    'instance_host': 'dev',
    'mac': 'fa:16:3e:90:4b:6f',
    'host': '5b85dd1e7796f87bae6fea6dcb608cf907446eaa715de90d77dadb7b',
    'vnic_name': 'tap704159b2-8b',
    'instance_id': '685243d6-34f1-4a39-8446-87b018ada8d9'}

TRANS_NETWORK_TAP = {
    'instance_host': 'dev',
    'mac': 'fa:16:3e:90:4b:6f',
    'host': '5b85dd1e7796f87bae6fea6dcb608cf907446eaa715de90d77dadb7b',
    'vnic_name': 'tap704159b2-8b',
    'instance_id': '685243d6-34f1-4a39-8446-87b018ada8d9',
    'project_id': '4480c638-3169-11e6-91de-a3bd3a7d3afb',
    'interface_id': '2f58a438-3169-11e6-b36c-bfe1fa3241fe',
    'user_id': '576808d8-3169-11e6-992b-5f931fc671df'}

CEIL_NETWORK_FLOATING = {
    'status': 'ACTIVE',
    'router_id': 'a46530af-c5ba-4ba2-aa59-e2de4393151d',
    'floating_network_id': '823daec0-b000-446b-9539-20f7463775c3',
    'fixed_ip_address': '10.0.0.6',
    'floating_ip_address': '172.24.4.9',
    'port_id': '65005c79-2ab0-46d4-8ab8-f3044ec47418'}

TRANS_NETWORK_FLOATING = {
    'project_id': '4480c638-3169-11e6-91de-a3bd3a7d3afb',
    'user_id': '576808d8-3169-11e6-992b-5f931fc671df',
    'floatingip_id': '2f58a438-3169-11e6-b36c-bfe1fa3241fe',
    'status': 'ACTIVE',
    'router_id': 'a46530af-c5ba-4ba2-aa59-e2de4393151d',
    'floating_network_id': '823daec0-b000-446b-9539-20f7463775c3',
    'fixed_ip_address': '10.0.0.6',
    'floating_ip_address': '172.24.4.9',
    'port_id': '65005c79-2ab0-46d4-8ab8-f3044ec47418'}


class CeilometerTransformerTest(tests.TestCase):
    def generate_ceilometer_resource(self, data):
        resource = t_transformers.ClassWithAttr({
            'project_id': '4480c638-3169-11e6-91de-a3bd3a7d3afb',
            'resource_id': '2f58a438-3169-11e6-b36c-bfe1fa3241fe',
            'user_id': '576808d8-3169-11e6-992b-5f931fc671df'})
        resource.metadata = copy.deepcopy(data)
        return resource

    def test_strip_ceilometer_compute(self):
        resource = self.generate_ceilometer_resource(CEIL_COMPUTE)
        t_test = ceilometer.CeilometerTransformer()
        result = t_test.strip_resource_data('compute', resource)
        self.assertEqual(TRANS_COMPUTE, result)

    def test_strip_ceilometer_volume(self):
        resource = self.generate_ceilometer_resource(CEIL_VOLUME)
        t_test = ceilometer.CeilometerTransformer()
        result = t_test.strip_resource_data('volume', resource)
        self.assertEqual(TRANS_VOLUME, result)

    def test_strip_ceilometer_image(self):
        resource = self.generate_ceilometer_resource(CEIL_IMAGE)
        t_test = ceilometer.CeilometerTransformer()
        result = t_test.strip_resource_data('image', resource)
        self.assertEqual(TRANS_IMAGE, result)

    def test_strip_ceilometer_network_tap(self):
        resource = self.generate_ceilometer_resource(CEIL_NETWORK_TAP)
        t_test = ceilometer.CeilometerTransformer()
        result = t_test.strip_resource_data('network_tap', resource)
        self.assertEqual(TRANS_NETWORK_TAP, result)

    def test_strip_ceilometer_network_floating(self):
        resource = self.generate_ceilometer_resource(CEIL_NETWORK_FLOATING)
        t_test = ceilometer.CeilometerTransformer()
        result = t_test.strip_resource_data('network_floating', resource)
        self.assertEqual(TRANS_NETWORK_FLOATING, result)



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import abc

from oslo_config import cfg
import six
from stevedore import driver

from cloudkitty import transformer
from cloudkitty import utils as ck_utils


CONF = cfg.CONF

METRICS_CONF = ck_utils.get_metrics_conf(CONF.collect.metrics_conf)

COLLECTORS_NAMESPACE = 'cloudkitty.collector.backends'


def get_collector(transformers=None):
    if not transformers:
        transformers = transformer.get_transformers()
    collector_args = {
        'period': METRICS_CONF['period'],
        'transformers': transformers}
    collector = driver.DriverManager(
        COLLECTORS_NAMESPACE,
        METRICS_CONF['collector'],
        invoke_on_load=True,
        invoke_kwds=collector_args).driver
    return collector


def get_collector_metadata():
    """Return dict of metadata.

    Results are based on enabled collector and services in CONF.
    """
    transformers = transformer.get_transformers()
    collector = driver.DriverManager(
        COLLECTORS_NAMESPACE, METRICS_CONF['collector'],
        invoke_on_load=False).driver
    metadata = {}
    for service in METRICS_CONF['services']:
        metadata[service] = collector.get_metadata(service, transformers)
    return metadata


class TransformerDependencyError(Exception):
    """Raised when a collector can't find a mandatory transformer."""

    def __init__(self, collector, transformer):
        super(TransformerDependencyError, self).__init__(
            "Transformer '%s' not found, but required by %s" % (transformer,
                                                                collector))
        self.collector = collector
        self.transformer = transformer


class NoDataCollected(Exception):
    """Raised when the collection returned no data.

    """

    def __init__(self, collector, resource):
        super(NoDataCollected, self).__init__(
            "Collector '%s' returned no data for resource '%s'" % (
                collector, resource))
        self.collector = collector
        self.resource = resource


@six.add_metaclass(abc.ABCMeta)
class BaseCollector(object):
    collector_name = None
    dependencies = []

    def __init__(self, transformers, **kwargs):
        try:
            self.transformers = transformers
            self.period = kwargs['period']
        except IndexError as e:
            raise ValueError("Missing argument (%s)" % e)

        self._check_transformers()

    def _check_transformers(self):
        """Check for transformer prerequisites

        """
        for dependency in self.dependencies:
            if dependency not in self.transformers:
                raise TransformerDependencyError(self.collector_name,
                                                 dependency)

    @staticmethod
    def last_month():
        month_start = ck_utils.get_month_start()
        month_end = ck_utils.get_month_end()
        start_ts = ck_utils.dt2ts(month_start)
        end_ts = ck_utils.dt2ts(month_end)
        return start_ts, end_ts

    @staticmethod
    def current_month():
        month_start = ck_utils.get_month_start()
        return ck_utils.dt2ts(month_start)

    @classmethod
    def _res_to_func(cls, resource_name):
        trans_resource = 'get_'
        trans_resource += resource_name.replace('.', '_')
        return trans_resource

    @classmethod
    def get_metadata(cls, resource_name, transformers):
        """Return metadata about collected resource as a dict.

           Dict object should contain:
                - "metadata": available metadata list,
                - "unit": collected quantity unit
        """
        return {"metadata": [], "unit": "undefined"}

    def retrieve(self,
                 resource,
                 start,
                 end=None,
                 project_id=None,
                 q_filter=None):
        trans_resource = self._res_to_func(resource)
        if not hasattr(self, trans_resource):
            raise NotImplementedError(
                "No method found in collector '%s' for resource '%s'."
                % (self.collector_name, resource))
        func = getattr(self, trans_resource)
        return func(start, end, project_id, q_filter)



# -*- coding: utf-8 -*-
# Copyright 2017 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Luka Peschke
#
import datetime
import decimal
import json

from gnocchiclient import client as gclient
from gnocchiclient import exceptions as gexceptions
from keystoneauth1 import loading as ks_loading
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import uuidutils
import six

from cloudkitty.storage.hybrid.backends import BaseHybridBackend
from cloudkitty.transformer import gnocchi as gtransformer
import cloudkitty.utils as ck_utils


LOG = logging.getLogger(__name__)

CONF = cfg.CONF

CONF.import_opt('period', 'cloudkitty.collector', 'collect')

GNOCCHI_STORAGE_OPTS = 'storage_gnocchi'
gnocchi_storage_opts = [
    cfg.StrOpt('archive_policy_name',
               default='rating',
               help='Gnocchi storage archive policy name.'),
    # The archive policy definition MUST include the collect period granularity
    cfg.StrOpt('archive_policy_definition',
               default='[{"granularity": '
                       + six.text_type(CONF.collect.period) +
                       ', "timespan": "90 days"}, '
                       '{"granularity": 86400, "timespan": "360 days"}, '
                       '{"granularity": 2592000, "timespan": "1800 days"}]',
               help='Gnocchi storage archive policy definition.'), ]
CONF.register_opts(gnocchi_storage_opts, GNOCCHI_STORAGE_OPTS)

ks_loading.register_session_conf_options(
    CONF,
    GNOCCHI_STORAGE_OPTS)
ks_loading.register_auth_conf_options(
    CONF,
    GNOCCHI_STORAGE_OPTS)

METRICS_CONF = ck_utils.get_metrics_conf(CONF.collect.metrics_conf)

RESOURCE_TYPE_NAME_ROOT = 'rating_service_'


class DecimalJSONEncoder(json.JSONEncoder):
    """Wrapper class to handle decimal.Decimal objects in json.dumps()."""
    def default(self, obj):
        if isinstance(obj, decimal.Decimal):
            return float(obj)
        return super(DecimalJSONEncoder, self).default(obj)


class UnknownResourceType(Exception):
    """Exception raised when an unknown resource type is encountered"""

    def __init__(self, resource_type):
        super(UnknownResourceType, self).__init__(
            'Unknown resource type {}'.format(resource_type)
        )


class GnocchiStorage(BaseHybridBackend):
    """Gnocchi backend for hybrid storage.

    """

    # NOTE(lukapeschke): List taken directly from gnocchi code
    invalid_attribute_names = [
        "id", "type", "metrics",
        "revision", "revision_start", "revision_end",
        "started_at", "ended_at",
        "user_id", "project_id",
        "created_by_user_id", "created_by_project_id", "get_metric",
        "creator",
    ]

    @staticmethod
    def _get_service_metrics(service_name):
        metrics = METRICS_CONF['services_metrics'][service_name]
        metric_list = ['price']
        for metric in metrics:
            metric_list.append(list(metric.keys())[0])
        return metric_list

    def _init_resource_types(self):
        transformer = gtransformer.GnocchiTransformer()
        services = METRICS_CONF['services']
        for service in services:
            service_dict = dict()
            service_dict['attributes'] = list()
            for attribute in transformer.get_metadata(service):
                if attribute not in self.invalid_attribute_names:
                    service_dict['attributes'].append(attribute)
            service_dict['required_attributes'] = [
                'resource_id',
                'unit',
            ]
            try:
                service_dict['metrics'] = self._get_service_metrics(service)
            except KeyError:
                LOG.warning(
                    'No metrics configured for service {}'.format(service))
                service_dict['metrics'] = list()
            service_dict['name'] = RESOURCE_TYPE_NAME_ROOT + service
            service_dict['qty_metric'] \
                = list(METRICS_CONF['metrics_units'][service].keys())[0]
            self._resource_type_data[service] = service_dict

    def _get_res_type_dict(self, res_type):
        res_type_data = self._resource_type_data.get(res_type, None)
        if not res_type_data:
            return None
        attribute_dict = dict()
        for attribute in res_type_data['attributes']:
            attribute_dict[attribute] = {
                'required': False,
                'type': 'string',
            }
        for attribute in res_type_data['required_attributes']:
            attribute_dict[attribute] = {
                'required': True,
                'type': 'string',
            }
        return {
            'name': res_type_data['name'],
            'attributes': attribute_dict,
        }

    def _create_resource(self, res_type, tenant_id, data):
        res_type_data = self._resource_type_data.get(res_type, None)
        if not res_type_data:
            raise UnknownResourceType(
                "Unknown resource type '{}'".format(res_type))

        res_dict = {
            'id': data['resource_id'],
            'resource_id': data['resource_id'],
            'project_id': tenant_id,
            'user_id': data['user_id'],
            'unit': data['unit'],
        }
        for attr in res_type_data['attributes']:
            res_dict[attr] = data.get(attr, None) or 'None'
            if isinstance(res_dict[attr], decimal.Decimal):
                res_dict[attr] = float(res_dict[attr])

        created_metrics = [
            self._conn.metric.create({
                'name': metric,
                'archive_policy_name':
                    CONF.storage_gnocchi.archive_policy_name,
            }) for metric in res_type_data['metrics']
        ]

        metrics_dict = dict()
        for metric in created_metrics:
            metrics_dict[metric['name']] = metric['id']
        res_dict['metrics'] = metrics_dict
        try:
            return self._conn.resource.create(res_type_data['name'], res_dict)
        except gexceptions.ResourceAlreadyExists:
            res_dict['id'] = uuidutils.generate_uuid()
            return self._conn.resource.create(res_type_data['name'], res_dict)

    def _get_resource(self, resource_type, resource_id):
        try:
            resource_name = self._resource_type_data[resource_type]['name']
        except KeyError:
            raise UnknownResourceType(
                "Unknown resource type '{}'".format(resource_type))
        try:
            return self._conn.resource.get(resource_name, resource_id)
        except gexceptions.ResourceNotFound:
            return None

    def _find_resource(self, resource_type, resource_id):
        try:
            resource_type = self._resource_type_data[resource_type]['name']
        except KeyError:
            raise UnknownResourceType(
                "Unknown resource type '{}'".format(resource_type))
        query = {
            '=': {
                'resource_id': resource_id,
            }
        }
        try:
            return self._conn.resource.search(
                resource_type=resource_type, query=query, limit=1)[0]
        except IndexError:
            return None

    def _create_resource_type(self, resource_type):
        res_type = self._resource_type_data.get(resource_type, None)
        if not res_type:
            return None
        res_type_dict = self._get_res_type_dict(resource_type)
        try:
            output = self._conn.resource_type.create(res_type_dict)
        except gexceptions.ResourceTypeAlreadyExists:
            output = None
        return output

    def _get_resource_type(self, resource_type):
        res_type = self._resource_type_data.get(resource_type, None)
        if not res_type:
            return None
        return self._conn.resource_type.get(res_type['name'])

    def __init__(self, **kwargs):
        super(GnocchiStorage, self).__init__(**kwargs)
        self.auth = ks_loading.load_auth_from_conf_options(
            CONF,
            GNOCCHI_STORAGE_OPTS)
        self.session = ks_loading.load_session_from_conf_options(
            CONF,
            GNOCCHI_STORAGE_OPTS,
            auth=self.auth)
        self._conn = gclient.Client('1', session=self.session)
        self._archive_policy_name = (
            CONF.storage_gnocchi.archive_policy_name)
        self._archive_policy_definition = json.loads(
            CONF.storage_gnocchi.archive_policy_definition)
        self._period = CONF.collect.period
        if "period" in kwargs:
            self._period = kwargs["period"]
        self._measurements = dict()
        self._resource_type_data = dict()
        self._init_resource_types()

    def commit(self, tenant_id, state):
        if not self._measurements.get(tenant_id, None):
            return
        commitable_measurements = dict()
        for metrics in self._measurements[tenant_id].values():
            for metric_id, measurements in metrics.items():
                if measurements:
                    measures = list()
                    for measurement in measurements:
                        measures.append(
                            {
                                'timestamp': state,
                                'value': measurement,
                            }
                        )
                    commitable_measurements[metric_id] = measures
        if commitable_measurements:
            self._conn.metric.batch_metrics_measures(commitable_measurements)
        del self._measurements[tenant_id]

    def init(self):
        try:
            self._conn.archive_policy.get(self._archive_policy_name)
        except gexceptions.ArchivePolicyNotFound:
            ck_archive_policy = {}
            ck_archive_policy['name'] = self._archive_policy_name
            ck_archive_policy['back_window'] = 0
            ck_archive_policy['aggregation_methods'] \
                = ['std', 'count', 'min', 'max', 'sum', 'mean']
            ck_archive_policy['definition'] = self._archive_policy_definition
            self._conn.archive_policy.create(ck_archive_policy)
        for service in self._resource_type_data.keys():
            try:
                self._get_resource_type(service)
            except gexceptions.ResourceTypeNotFound:
                self._create_resource_type(service)

    def get_total(self, begin=None, end=None, tenant_id=None,
                  service=None, groupby=None):
        # Query can't be None if we don't specify a resource_id
        query = {}
        if tenant_id:
            query['='] = {"project_id": tenant_id}
        measures = self._conn.metric.aggregation(
            metrics='price', query=query,
            start=begin, stop=end,
            aggregation='sum',
            granularity=self._period,
            needed_overlap=0)
        rate = sum(measure[2] for measure in measures) if len(measures) else 0
        return [{
            'begin': begin,
            'end': end,
            'rate': rate,
        }]

    def _append_measurements(self, resource, data, tenant_id):
        if not self._measurements.get(tenant_id, None):
            self._measurements[tenant_id] = {}
        measurements = self._measurements[tenant_id]
        if not measurements.get(resource['id'], None):
            measurements[resource['id']] = {
                key: list() for key in resource['metrics'].values()
            }
        for metric_name, metric_id in resource['metrics'].items():
            measurement = data.get(metric_name, None)
            if measurement is not None:
                measurements[resource['id']][metric_id].append(
                    float(measurement)
                    if isinstance(measurement, decimal.Decimal)
                    else measurement)

    def append_time_frame(self, res_type, frame, tenant_id):
        flat_frame = ck_utils.flat_dict(frame)
        resource = self._find_resource(res_type, flat_frame['resource_id'])
        if not resource:
            resource = self._create_resource(res_type, tenant_id, flat_frame)
        self._append_measurements(resource, flat_frame, tenant_id)

    def get_tenants(self, begin, end):
        query = {'like': {'type': RESOURCE_TYPE_NAME_ROOT + '%'}}
        r = self._conn.metric.aggregation(
            metrics='price',
            query=query,
            start=begin,
            stop=end,
            aggregation='sum',
            granularity=self._period,
            needed_overlap=0,
            groupby='project_id')
        projects = list()
        for measures in r:
            projects.append(measures['group']['project_id'])
        return projects

    @staticmethod
    def _get_time_query(start, end, resource_type, tenant_id=None):
        query = {'and': [{
            'or': [
                {'=': {'ended_at': None}},
                {'<=': {'ended_at': end}}
            ]
        },
            {'>=': {'started_at': start}},
            {'=': {'type': resource_type}},
        ]
        }
        if tenant_id:
            query['and'].append({'=': {'project_id': tenant_id}})
        return query

    def _get_resources(self, resource_type, start, end, tenant_id=None):
        """Returns the resources of the given type in the given period"""
        return self._conn.resource.search(
            resource_type=resource_type,
            query=self._get_time_query(start, end, resource_type, tenant_id),
            details=True)

    def _format_frame(self, res_type, resource, desc, measure, tenant_id):
        res_type_info = self._resource_type_data.get(res_type, None)
        if not res_type_info:
            return dict()

        start = measure[0]
        stop = start + datetime.timedelta(seconds=self._period)

        # Getting price
        price = decimal.Decimal(measure[2])
        price_dict = {'price': float(price)}

        # Getting vol
        if isinstance(res_type_info['qty_metric'], (str, unicode)):
            try:
                qty = self._conn.metric.get_measures(
                    resource['metrics'][res_type_info['qty_metric']],
                    aggregation='sum',
                    start=start, stop=stop,
                    refresh=True)[-1][2]
            except IndexError:
                qty = 0
        else:
            qty = res_type_info['qty_metric']
        vol_dict = {'qty': decimal.Decimal(qty), 'unit': resource['unit']}

        # Period
        period_dict = {
            'begin': ck_utils.dt2iso(start),
            'end': ck_utils.dt2iso(stop),
        }

        # Formatting
        res_dict = dict()
        res_dict['desc'] = desc
        res_dict['vol'] = vol_dict
        res_dict['rating'] = price_dict
        res_dict['tenant_id'] = tenant_id

        return {
            'usage': {res_type: [res_dict]},
            'period': period_dict,
        }

    def resource_info(self, resource_type, start, end, tenant_id=None):
        """Returns a dataframe for the given resource type"""
        try:
            res_type_info = self._resource_type_data.get(resource_type, None)
            resource_name = res_type_info['name']
        except (KeyError, AttributeError):
            raise UnknownResourceType(resource_type)
        attributes = res_type_info['attributes'] \
            + res_type_info['required_attributes']
        output = list()
        query = self._get_time_query(start, end, resource_name, tenant_id)
        measures = self._conn.metric.aggregation(
            metrics='price',
            resource_type=resource_name,
            query=query,
            start=start,
            stop=end,
            granularity=self._period,
            aggregation='sum',
            needed_overlap=0,
            groupby=['type', 'id'],
        )
        for resource_measures in measures:
            resource_desc = None
            resource = None
            for measure in resource_measures['measures']:
                if not resource_desc:
                    resource = self._get_resource(
                        resource_type, resource_measures['group']['id'])
                    if not resource:
                        continue
                    desc = {a: resource.get(a, None) for a in attributes}
                formatted_frame = self._format_frame(
                    resource_type, resource, desc, measure, tenant_id)
                output.append(formatted_frame)
        return output

    def get_time_frame(self, begin, end, **filters):
        tenant_id = filters.get('tenant_id', None)
        resource_types = [filters.get('res_type', None)]
        if not resource_types[0]:
            resource_types = self._resource_type_data.keys()
        output = list()
        for resource_type in resource_types:
            output += self.resource_info(resource_type, begin, end, tenant_id)
        return output



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Gauvain Pocentek
#
import decimal

import mock
from oslo_config import fixture as config_fixture
from oslotest import base
import testscenarios

from cloudkitty import collector
from cloudkitty import db
from cloudkitty.db import api as ck_db_api
from cloudkitty import rating


class FakeCollectorModule(collector.BaseCollector):
    collector_name = 'test_fake'
    dependencies = tuple()

    def __init__(self):
        super(FakeCollectorModule, self).__init__([], period=3600)


class FakeRatingModule(rating.RatingProcessorBase):
    module_name = 'fake'
    description = 'fake rating module'

    def __init__(self, tenant_id=None):
        super(FakeRatingModule, self).__init__()

    def quote(self, data):
        self.process(data)

    def process(self, data):
        for cur_data in data:
            cur_usage = cur_data['usage']
            for service in cur_usage:
                for entry in cur_usage[service]:
                    if 'rating' not in entry:
                        entry['rating'] = {'price': decimal.Decimal(0)}
        return data

    def reload_config(self):
        pass

    def notify_reload(self):
        pass


class TestCase(testscenarios.TestWithScenarios, base.BaseTestCase):
    scenarios = [
        ('sqlite', dict(db_url='sqlite:///'))
    ]

    def setUp(self):
        super(TestCase, self).setUp()
        self._conf_fixture = self.useFixture(config_fixture.Config())
        self.conf = self._conf_fixture.conf
        self.conf.set_override('connection', self.db_url, 'database')
        self.conn = ck_db_api.get_instance()
        migration = self.conn.get_migration()
        migration.upgrade('head')
        auth = mock.patch(
            'keystoneauth1.loading.load_auth_from_conf_options',
            return_value=dict())
        auth.start()
        self.auth = auth
        session = mock.patch(
            'keystoneauth1.loading.load_session_from_conf_options',
            return_value=dict())
        session.start()
        self.session = session

    def tearDown(self):
        db.get_engine().dispose()
        self.auth.stop()
        self.session.stop()
        super(TestCase, self).tearDown()
